
from vsg import token

from vsg.rules.whitespace_before_token import Rule

lTokens = []
lTokens.append(token.attribute_specification.is_keyword)


class rule_101(Rule):
    '''
    This rule checks for a single space before the **is** keyword.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       attribute coordinate of comp_1 : component   is (0.0, 17.5);

    **Fix**

    .. code-block:: vhdl

       attribute coordinate of comp_1 : component is (0.0, 17.5);

       attribute coordinate of comp_1 : component is (0.0, 17.5);
    '''
    def __init__(self):
        Rule.__init__(self, 'attribute_specification', '101', lTokens)




from vsg.rules import move_token_to_the_right_of_several_possible_tokens_if_it_exists_between_tokens

from vsg.token import block_statement as token

lAnchorTokens = []
lAnchorTokens.append(token.block_keyword)
lAnchorTokens.append(token.guard_close_parenthesis)

oToken = token.is_keyword

oStartToken = token.block_keyword
oEndToken = token.begin_keyword


class rule_003(move_token_to_the_right_of_several_possible_tokens_if_it_exists_between_tokens):
    '''
    This rule checks the **is** keyword is on the same line as the **block** keyword.

    **Violation**

    .. code-block:: vhdl

       block_label : block
       is

    **Fix**

    .. code-block:: vhdl

       block_labeel : block is
    '''

    def __init__(self):
        move_token_to_the_right_of_several_possible_tokens_if_it_exists_between_tokens.__init__(self, 'block', '003', oToken, lAnchorTokens, oStartToken, oEndToken, bInsertWhitespace=True)
        self.solution = 'Move block_simple_name next to *block* keyword'




from vsg.vhdlFile.extract import tokens


def get_tokens_bounded_by_tokens_if_token_is_between_them(oStart, oEnd, oToken, lAllObjects, oTokenMap):

    lReturn = []

    lStart, lEnd = oTokenMap.get_token_pair_indexes(oStart, oEnd)

    lIndexes = oTokenMap.get_token_indexes(oToken)

    for iStart, iEnd in zip(lStart, lEnd):
        for iIndex in lIndexes:
            if iStart < iIndex and iIndex < iEnd:
                lTemp = lAllObjects[iStart: iEnd + 1]
                iStartLine = oTokenMap.get_line_number_of_index(iStart)
                oToi = tokens.New(iStart, iStartLine, lTemp)
                oToi.set_token_value(iIndex - iStart)
                lReturn.append(oToi)

    return lReturn




from vsg.rules import align_tokens_in_region_between_tokens_skipping_lines_starting_with_tokens

from vsg import parser
from vsg import token

lAlign = []
lAlign.append(parser.comment)

lSkip = []
lSkip.append(parser.comment)


class rule_020(align_tokens_in_region_between_tokens_skipping_lines_starting_with_tokens):
    '''
    This rule checks for alignment of inline comments in the component declaration.

    Following extra configurations are supported:

    * :code:`separate_generic_port_alignment`.

    |configuring_keyword_alignment_rules_link|

    **Violation**

    .. code-block:: vhdl

       component my_component
           generic (
               g_width        : positive;  -- Data width
               g_output_delay : positive -- Delay at output
           );
           port (
               clk_i  : in std_logic; -- Input clock
               data_i : in std_logic;   -- Data input
               data_o : in std_logic -- Data output
           );
       end my_component;

    **Fix**

    .. code-block:: vhdl

       component my_component
           generic (
               g_width        : positive; -- Data width
               g_output_delay : positive  -- Delay at output
           );
           port (
               clk_i  : in std_logic; -- Input clock
               data_i : in std_logic; -- Data input
               data_o : in std_logic  -- Data output
           );
       end my_component;
    '''

    def __init__(self):
        align_tokens_in_region_between_tokens_skipping_lines_starting_with_tokens.__init__(self, 'component', '020', lAlign, token.component_declaration.component_keyword, token.component_declaration.end_keyword, lSkip)
        self.solution = 'Align identifer.'
        self.subphase = 2




from vsg.rules import token_case_with_prefix_suffix

from vsg import token

lTokens = []
lTokens.append(token.loop_statement.end_loop_label)


class rule_504(token_case_with_prefix_suffix):
    '''
    This rule checks the proper case of the end label on a loop statement.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

         end loop LABEL;
         end loop Label;

    **Fix**

    .. code-block:: vhdl

         end loop label;
         end loop label;
    '''

    def __init__(self):
        token_case_with_prefix_suffix.__init__(self, 'loop_statement', '504', lTokens)
        self.groups.append('case::label')




from vsg.rules import token_case

from vsg import token

lTokens = []
lTokens.append(token.process_statement.end_process_keyword)


class rule_009(token_case):
    '''
    This rule checks the **process** keyword has proper case in the **end process** line.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       end PROCESS proc_a;

    **Fix**

    .. code-block:: vhdl

       end process proc_a;
    '''

    def __init__(self):
        token_case.__init__(self, 'process', '009', lTokens)
        self.groups.append('case::keyword')




from vsg.rules import token_prefix

from vsg import token

lTokens = []
lTokens.append(token.loop_statement.loop_label)


class rule_600(token_prefix):
    '''
    This rule checks for valid prefixes on loop labels.
    The default prefix is *loop\_*.

    |configuring_prefix_and_suffix_rules_link|

    **Violation**

    .. code-block:: vhdl

       label : for index in 4 to 23 loop

    **Fix**

    .. code-block:: vhdl

       loop_label : for index in 4 to 23 loop
    '''

    def __init__(self):
        token_prefix.__init__(self, 'loop_statement', '600', lTokens)
        self.prefixes = ['loop_']
        self.solution = 'Loop labels'




from vsg import deprecated_rule


class rule_016(deprecated_rule.Rule):
    '''
    This rule has been renamed to `generic_map_002 <generic_map_rules.html#generic-map-002>`_.
    '''

    def __init__(self):
        deprecated_rule.Rule.__init__(self, 'instantiation', '016')
        self.message.append('Rule ' + self.unique_id + ' has been renamed to rule generic_map_002.')





from vsg import parser
from vsg import violation

from vsg.vhdlFile import utils
from vsg.rule_group import structure


class insert_tokens_right_of_token_if_it_does_not_exist_before_token(structure.Rule):
    '''
    Checks for the existence of a token and will insert it if it does not exist.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    insert_token : token object
       token to insert if it does not exist.

    anchor_token : token object type
       token to check if insert_token exists to the right of

    end_token : token object type
       token that bounds the search for the insert_token
    '''

    def __init__(self, name, identifier, insert_tokens, anchor_token, end_token):
        structure.Rule.__init__(self, name=name, identifier=identifier)
        self.insert_tokens = insert_tokens
        self.anchor_token = anchor_token
        self.end_token = end_token
        self.action = 'add'
        self.configuration.append('action')

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_tokens_bounded_by(self.anchor_token, self.end_token)

    def _analyze(self, lToi):
        if self.action == 'add':
            for oToi in lToi:
                lTokens = oToi.get_tokens()

                for oToken in lTokens:
                    if isinstance(oToken, type(self.insert_tokens[0])):
                        break
                else:
                    sSolution = self.action.capitalize() + ' ' + self.solution
                    self.add_violation(violation.New(oToi.get_line_number(), oToi, sSolution))
        else:
            for oToi in lToi:
                lTokens = oToi.get_tokens()
                for iToken, oToken in enumerate(lTokens):
                    if isinstance(oToken, type(self.insert_tokens[0])):
                        sSolution = self.action.capitalize() + ' ' + self.solution
                        dAction = {}
                        dAction['iStartIndex'] = iToken
                        dAction['iEndIndex'] = iToken + len(self.insert_tokens)
                        oViolation = violation.New(oToi.get_line_number(), oToi, sSolution)
                        oViolation.set_action(dAction)
                        self.add_violation(oViolation)
                        break

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        if self.action == 'add':
            lNewTokens = []
            lNewTokens.append(lTokens[0])
            if isinstance(lTokens[1], parser.whitespace) and isinstance(lTokens[2], parser.semicolon):
                lNewTokens.append(lTokens[1])
                lNewTokens.extend(self.insert_tokens)
                lNewTokens.extend(lTokens[2:])
            else:
                lNewTokens.append(parser.whitespace(' '))
                lNewTokens.extend(self.insert_tokens)
                lNewTokens.extend(lTokens[1:])
        else:
            dAction = oViolation.get_action()
            lNewTokens = lTokens[:dAction['iStartIndex']]
            lNewTokens.extend(lTokens[dAction['iEndIndex']:])
            lNewTokens = utils.remove_consecutive_whitespace_tokens(lNewTokens)
        oViolation.set_tokens(lNewTokens)




from vsg.token import variable_assignment_statement as token

from vsg.vhdlFile.classify import conditional_variable_assignment
from vsg.vhdlFile.classify import selected_variable_assignment
from vsg.vhdlFile.classify import simple_variable_assignment

from vsg.vhdlFile import utils


def detect(iToken, lObjects):
    '''
    variable_assignment_statement ::=
        [ label : ] simple_variable_assignment
      | [ label : ] conditional_variable_assignment
      | [ label : ] selected_variable_assignment
    '''
    iCurrent = iToken
    if selected_variable_assignment.detect(iToken, lObjects):
        iCurrent = utils.tokenize_label(iCurrent, lObjects, token.label, token.label_colon)
        iCurrent = selected_variable_assignment.classify(iCurrent, lObjects)

    elif conditional_variable_assignment.detect(iToken, lObjects):
        iCurrent = utils.tokenize_label(iCurrent, lObjects, token.label, token.label_colon)
        iCurrent = conditional_variable_assignment.classify(iCurrent, lObjects)

    elif simple_variable_assignment.detect(iToken, lObjects):
        iCurrent = utils.tokenize_label(iCurrent, lObjects, token.label, token.label_colon)
        iCurrent = simple_variable_assignment.classify(iCurrent, lObjects)

    return iCurrent




from vsg import parser
from vsg import violation

from vsg.rules import utils
from vsg.rule_group import structure


class rule_011(structure.Rule):
    '''
    This rule checks for in-line comments and moves them to the line above.
    The indent of the comment will be set to the indent of the current line.

    .. NOTE:: This rule is disabled by default.

    **Violation**

    .. code-block:: vhdl

       a <= b; -- Assign signal

    **Fix**

    .. code-block:: vhdl

       -- Assign signal
       a <= b;
    '''

    def __init__(self):
        structure.Rule.__init__(self, name='comment', identifier='011')
        self.solution = 'Move inline comment to previous line.'
        self.disable = True
        self.lTokens = [parser.comment]

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_line_which_includes_tokens(self.lTokens)

    def _analyze(self, lToi):
        for oToi in lToi:
            if line_starts_with_comment(oToi):
                continue
            self._check_for_inline_comment(oToi)

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()

        lTemp = lTokens[dAction['iToken']:]
        lTemp.append(parser.carriage_return())
        lTemp.extend(lTokens[:dAction['iToken']])
#        lTemp[dAction['index']].set_indent(dAction['indent'])

        oViolation.set_tokens(lTemp)

    def _check_for_inline_comment(self, oToi):
        iLine, lTokens = utils.get_toi_parameters(oToi)
        iToken = len(lTokens) - 1
        if line_has_inline_comment(lTokens, iToken):
            dAction = create_action(lTokens, iToken)
            oViolation = create_violation(iLine, oToi, self.solution, dAction)
            self.add_violation(oViolation)


def line_starts_with_comment(oToi):
    if utils.does_line_start_with_comment(oToi.get_tokens()):
        return True
    return False


def line_has_inline_comment(lTokens, iToken):
    if isinstance(lTokens[iToken], parser.comment):
        return True
    return False


def create_action(lTokens, iToken):
    dAction = {}
    if isinstance(lTokens[iToken - 1], parser.whitespace):
        dAction['iToken'] = iToken - 1
        dAction['index'] = 1
    else:
        dAction['iToken'] = iToken
        dAction['index'] = 0
#    dAction['indent'] = utils.get_indent_of_line(lTokens)
    return dAction


def create_violation(iLine, oToi, solution, dAction):
    oViolation = violation.New(iLine, oToi, solution)
    oViolation.set_action(dAction)
    return oViolation




import subprocess
import os

sVersion = '3.13.0'


def print_version(oCommandLineArguments):

    if (oCommandLineArguments.version):

        sVersion, sShaNum = get_version_info()

        print('VHDL Style Guide (VSG) version: ' + str(sVersion))

        print('Git commit SHA: ' + str(sShaNum))

        exit(0)


def get_version_info():

    sVersion = '3.13.0'

    if reporting_from_zip_file():
        return sVersion + '+zip.file', 'Unknown.  Installed via zip file.'

    if installed():
        from vsg import version_info
        return version_info.sVersion, version_info.sShaNum

    if reporting_from_git_repo():
        sReturnPath = os.getcwd()
        sPath = os.path.dirname(__file__)
        os.chdir(sPath)
        lActual = subprocess.check_output(['git', 'describe', '--tags'])
        lActual = str(lActual.decode('utf-8')).split('\n')
        lVersion = lActual[0].split('-')
        try:
            sVersion = str(lVersion[0]) + '.dev' + str(lVersion[1])
            sShaNum = str(lVersion[-1][1:])
        except IndexError:
            sVersion = str(lVersion[0])
            lActual = subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD'])
            lActual = str(lActual.decode('utf-8')).split('\n')
            sShaNum = str(lActual[0])
        os.chdir(sReturnPath)
        return sVersion, sShaNum


def reporting_from_git_repo():
    sPath = os.path.dirname(__file__)
    if os.path.isdir(os.path.join(sPath, '..', '.git')):
        return True
    return False


def installed():
    if not reporting_from_git_repo() and version_info_file_exists():
        return True
    return False


def reporting_from_zip_file():
    if not reporting_from_git_repo() and not installed():
        return True
    return False


def version_info_file_exists():
    sVersionInfoFileName = os.path.join('vsg', 'version_info.py')
    return os.path.exists(sVersionInfoFileName)




from vsg.rules import blank_line_above_line_starting_with_token

from vsg import token

lTokens = []
lTokens.append(token.package_declaration.end_keyword)


class rule_012(blank_line_above_line_starting_with_token):
    '''
    This rule checks for blank lines or comments above the **end package** keyword.

    |configuring_blank_lines_link|

    **Violation**

    .. code-block:: vhdl

         constant depth : integer := 512;
       end package FIFO_PKG;

    **Fix**

    .. code-block:: vhdl

         constant depth : integer := 512;

       end package FIFO_PKG;
    '''

    def __init__(self):
        blank_line_above_line_starting_with_token.__init__(self, 'package', '012', lTokens)




from vsg.rules import insert_token_left_of_token_if_it_does_not_exist_between_tokens_using_value_from_token

from vsg.token import architecture_body as token


class rule_024(insert_token_left_of_token_if_it_does_not_exist_between_tokens_using_value_from_token):
    '''
    This rule checks for the architecture name in the **end architecture** statement.
    It is clearer to the reader to state which architecture the end statement is closing.

    |configuring_optional_items_link|

    **Violation**

    .. code-block:: vhdl

       end architecture;

    **Fix**

    .. code-block:: vhdl

       end architecture architecture_name;
    '''

    def __init__(self):
        insert_token_left_of_token_if_it_does_not_exist_between_tokens_using_value_from_token.__init__(self, 'architecture', '024', token.architecture_simple_name, token.semicolon, token.end_keyword, token.semicolon, token.identifier)
        self.solution = 'architecture simple name'
        self.groups.append('structure::optional')




from vsg import parser


class when_keyword(parser.keyword):
    '''
    unique_id = conditional_waveforms : when_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class else_keyword(parser.keyword):
    '''
    unique_id = conditional_waveforms : else_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class waveform(parser.item):
    '''
    unique_id = conditional_waveforms : waveform
    '''

    def __init__(self, sString):
        parser.item.__init__(self, sString)


class condition(parser.item):
    '''
    unique_id = conditional_waveforms : condition
    '''

    def __init__(self, sString):
        parser.item.__init__(self, sString)





from vsg import severity
from vsg import utils


class Rule():

    def __init__(self, name=None, identifier=None):
        self.name = name
        self.identifier = identifier
        self.unique_id = str(name) + '_' + str(identifier)
        self.solution = None
        self.violations = []
        self.indentStyle = 'spaces'
        self.indentSize = 2
        self.phase = None
        self.subphase = 1
        self.disable = False
        self.fixable = True
        self.severity = severity.error('Error')
        self.debug = False
        self.dFix = {}
        self.dFix['violations'] = {}
        self.configuration = ['indentStyle', 'indentSize', 'phase', 'disable', 'fixable', 'severity']
        self.deprecated = False
        self.proposed = False
        self.groups = []

    def configure(self, oConfig):
        '''Configures attributes on rules using a dictionary of the following form:

            dConfiguration['rule'] = {}
            dConfiguration['rule']['xyz_001'] = {}
            dConfiguration['rule']['xyz_001']['disable'] = True
            dConfiguration['rule']['xyz_001']['solution'] = 'This is the new solution'
            dConfiguration['rule']['xyz_002'] = {}
            dConfiguration['rule']['xyz_002']['disable'] = False
            dConfiguration['rule']['global'] = {}
            dConfiguration['rule']['global']['indentSize'] = 4

          The rule:global dictionary will apply to all rules.
          Individual rule attributes can be modified with [self.name_self.identifier].
        '''
        if self.deprecated and self.unique_id in oConfig.dConfig['rule']:
            return self.print_output()

        dConfiguration = oConfig.dConfig
        configure_global_rule_attributes(self, oConfig)
        configure_group_rule_attributes(self, oConfig)
        configure_rule_attributes(self, oConfig)
        return []

    def has_violations(self):
        if len(self.violations) == 0:
            return False
        return True

    def _build_violation_dict_from_violation_object(self, violation):
        dViolation = {}
        dViolation['severity'] = {}
        dViolation['severity']['name'] = self.severity.name
        dViolation['severity']['type'] = self.severity.type
        dViolation['rule'] = self.get_unique_id()
        dViolation['lineNumber'] = str(violation.get_line_number())
        dViolation['solution'] = violation.get_solution()
        return dViolation

    def get_unique_id(self):
        return self.name + '_' + self.identifier

    def get_violations_at_linenumber(self, iLineNumber):
        '''
        Returns a list of formatted violations.

        Parameters:

          iLineNumber (integer)

        Returns: (list of dictionaries)
        '''
        lReturn = []

        for violation in self.violations:
            if violation.get_line_number() == iLineNumber:
                lReturn.append(self._build_violation_dict_from_violation_object(violation))

        return lReturn

    def get_violations(self):
        lReturn = []

        for violation in self.violations:
            lReturn.append(self._build_violation_dict_from_violation_object(violation))

        return lReturn

    def fix(self, oFile, dFixOnly=None):
        '''
        Applies fixes for any rule violations.
        '''
        if self.fixable:
            self.analyze(oFile)
            self._print_debug_message('Fixing rule: ' + self.unique_id)
            self._filter_out_fix_only_violations(dFixOnly)
            for oViolation in self.violations[::-1]:
                self._fix_violation(oViolation)
            oFile.update(self.violations)
            self.clear_violations()

    def _filter_out_fix_only_violations(self, dFixOnly):
        if dFixOnly is None:
            return

        try:
            if 'all' in dFixOnly['fix']['rule'][self.unique_id]:
                return
        except KeyError:
            self.violations = []

        lTemp = []
        for oViolation in self.violations:
            if oViolation.get_line_number() in dFixOnly['fix']['rule'][self.unique_id]:
                lTemp.append(oViolation)
        self.violations = lTemp

    def add_violation(self, violation):
        '''
        Adds a linenumber to a violations list.
        '''
        if not violation.has_code_tag(self.unique_id):
            self.violations.append(violation)

    def analyze(self, oFile):
        '''
        Performs the analysis.
        '''
        self._print_debug_message('Analyzing rule: ' + self.unique_id)
        lToi = self._get_tokens_of_interest(oFile)
        self._analyze(lToi)

    def get_configuration(self):
        '''
        Returns a dictionary of every configurable attribute of the rule.
        '''
        dConfig = {}
        for sParameter in self.configuration:
            dConfig[sParameter] = getattr(self, sParameter)
        dConfig['severity'] = self.severity.name
        return dConfig

    def _get_solution(self, iLineNumber):
        '''
        Returns the solution for a violation.
        '''
        return utils.get_violation_solution_at_line_number(self.violations, iLineNumber)

    def clear_violations(self):
        self.violations = []

    def set_debug(self):
        '''
        This method sets the debug attribute to True.
        '''
        self.debug = True

    def _print_debug_message(self, sString):
        '''
        Prints a debug message to stdio.

        Parameters:

          sString : (string)

        Returns: Nothing
        '''
        if self.debug:
            print('INFO: ' + sString)

    def _fix_violation(self, oViolation):
        '''
        Default if rule is not fixable
        '''
        return

    def _sort_violations(self):
        '''
        Sorts self.violations based on the start index of each violation.
        Sort order is increasing start index.
        '''
        if not self.violations:
            return
        lNewViolations = []
        lNewViolations.append(self.violations[0])
        for oViolation in self.violations[1:]:
            iStartIndex = oViolation.get_start_index()
            for iInsertIndex, oNewViolation in enumerate(lNewViolations):
                iNewIndex = oNewViolation.get_start_index()
                if iStartIndex < iNewIndex:
                    lNewViolations.insert(iInsertIndex, oViolation)
                    break
            else:
                lNewViolations.append(oViolation)
        self.violations = lNewViolations


def configure_group_rule_attributes(self, oConfig):
    '''
    Updates rule attributes based on configuration input files
    '''
    try:
        for sGroupName in oConfig.dConfig['rule']['group']:
            if sGroupName in self.groups:
                configure_attribute(self, oConfig, sGroupName)
    except KeyError:
        pass


def configure_attribute(self, oConfig, sGroupName):
    try:
        for sAttributeName in oConfig.dConfig['rule']['group'][sGroupName]:
            if sAttributeName == 'severity':
                self.severity = oConfig.severity_list.get_severity_named(oConfig.dConfig['rule']['group'][sGroupName]['severity'])
            elif sAttributeName in self.__dict__:
                self.__dict__[sAttributeName] = oConfig.dConfig['rule']['group'][sGroupName][sAttributeName]
    except KeyError:
        pass


def configure_global_rule_attributes(self, oConfig):
    '''
    Updates rule attributes based on configuration input files
    '''
    try:
        for sAttributeName in oConfig.dConfig['rule']['global']:
            if sAttributeName == 'severity':
                self.severity = oConfig.severity_list.get_severity_named(oConfig.dConfig['rule']['global']['severity'])
            elif sAttributeName in self.configuration:
                self.__dict__[sAttributeName] = oConfig.dConfig['rule']['global'][sAttributeName]
    except KeyError:
        pass


def configure_rule_attributes(self, oConfig):
    '''
    Updates rule attributes based on configuration input files
    '''
    try:
        for sAttributeName in oConfig.dConfig['rule'][self.get_unique_id()]:
            if sAttributeName == 'severity':
                self.severity = oConfig.severity_list.get_severity_named(oConfig.dConfig['rule'][self.get_unique_id()]['severity'])
            elif sAttributeName in self.__dict__:
                self.__dict__[sAttributeName] = oConfig.dConfig['rule'][self.get_unique_id()][sAttributeName]
    except KeyError:
        pass




from vsg.rules import align_tokens_in_region_between_tokens as Rule

from vsg import token

lAlign = []
lAlign.append(token.element_association.assignment)

oStart = token.concurrent_simple_signal_assignment.assignment
oEnd = token.concurrent_simple_signal_assignment.semicolon


class rule_400(Rule):
    '''
    This rule checks the alignment the => operator in record aggregates.

    |configuring_keyword_alignment_rules_link|

    **Violation**

    .. code-block:: vhdl

       interface <= (
                     write_words => 12,
                     read_words => 32
                     address => 57
                    );

    **Fix**

    .. code-block:: vhdl

       interface <= (
                     write_words => 12,
                     read_words  => 32
                     address     => 57
                    );
    '''

    def __init__(self):
        Rule.__init__(self, 'concurrent', '400', lAlign, oStart, oEnd)
        self.phase = 5
        self.subphase = 3
        self.solution = 'Align =>'
        self.separate_generic_port_alignment = False
        self.comment_line_ends_group = False
        self.blank_line_ends_group = False
        self.configuration.remove('separate_generic_port_alignment')
        self.bIncludeTillBeginningOfLine = True




from vsg import token

from vsg.rules.whitespace_before_token import Rule

lTokens = []
lTokens.append(token.loop_statement.loop_keyword)


class rule_102(Rule):
    '''
    This rule checks for a single space before the **loop** keyword.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

      for x in (0 to 30)loop
      for x in (0 to 30)         loop

    **Fix**

    .. code-block:: vhdl

      for x in (0 to 30) loop
      for x in (0 to 30) loop
    '''
    def __init__(self):
        Rule.__init__(self, 'loop_statement', '102', lTokens)




from vsg import token

from vsg.rules import check
from vsg.rules import fix
from vsg.rules import tokens_of_interest as toi

from vsg.rule_group import structure

from vsg.vhdlFile import utils


class multiline_procedure_call_structure(structure.Rule):
    '''
    This rule checks the structure of multiline constraints.
    '''

    def __init__(self, name, identifier):
        structure.Rule.__init__(self, name=name, identifier=identifier)
        self.phase = 1
        self.lTokenPairs = None

        self.first_open_paren = 'ignore'
        self.configuration.append('first_open_paren')
        self.last_close_paren = 'ignore'
        self.configuration.append('last_close_paren')
        self.association_list_comma = 'ignore'
        self.configuration.append('association_list_comma')
        self.association_element = 'ignore'
        self.configuration.append('association_element')

    def _get_tokens_of_interest(self, oFile):
        return toi.get_tokens_bounded_by(self.lTokenPairs, oFile)

    def _analyze(self, lToi):
        for oToi in lToi:

            _check_first_open_paren(self, oToi)
            _check_last_close_paren(self, oToi)
            _check_association_list_comma(self, oToi)
            _check_association_element(self, oToi)

        self._sort_violations()

    def _fix_violation(self, oViolation):
        fix.fix_violation(oViolation)


def _check_first_open_paren(self, oToi):

    check.add_new_line_and_remove_new_line(self, oToi, self.first_open_paren, token.procedure_call.open_parenthesis)


def _check_last_close_paren(self, oToi):

    check.add_new_line_and_remove_new_line(self, oToi, self.last_close_paren, token.procedure_call.close_parenthesis)


def _check_association_list_comma(self, oToi):

    check.add_new_line_and_remove_new_line(self, oToi, self.association_list_comma, token.association_list.comma)


def _check_association_element(self, oToi):

    check.add_new_line_and_remove_new_line(self, oToi, self.association_element, token.association_element.formal_part)





from vsg.rules import multiline_array_alignment as Rule

from vsg import token

lTokenPairs = []
lTokenPairs.append([token.concurrent_simple_signal_assignment.assignment, token.concurrent_simple_signal_assignment.semicolon])


class rule_401(Rule):
    '''
    This rule checks the alignment of multiline concurrent simple signal assignments that contain arrays.

    |configuring_multiline_indent_rules_link|

    **Violation**

    .. code-block:: vhdl

       wr_data <=
       (
                0,
            65535,
            32768
         );

    **Fix**

    .. code-block:: vhdl

       wr_data <=
       (
         0,
         65535,
         32768
       );
    '''

    def __init__(self):
        Rule.__init__(self, 'concurrent', '401', lTokenPairs)




from vsg.rules.whitespace_between_token_pairs import Rule

from vsg import token

lTokens = []
lTokens.append([token.concurrent_simple_signal_assignment.target, token.concurrent_simple_signal_assignment.assignment])
lTokens.append([token.concurrent_conditional_signal_assignment.target, token.concurrent_conditional_signal_assignment.assignment])
lTokens.append([token.concurrent_selected_signal_assignment.target, token.concurrent_selected_signal_assignment.assignment])


class rule_004(Rule):
    '''
    This rule checks for at least a single space before the **<=** operator.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       wr_en<= '0';

    **Fix**

    .. code-block:: vhdl

       wr_en <= '0';
    '''
    def __init__(self):
        Rule.__init__(self, 'concurrent', '004', lTokens)
        self.number_of_spaces = '>=1'




from vsg.rules import move_token as Rule

from vsg import token

oToken = token.port_map_aspect.close_parenthesis


class rule_004(Rule):
    '''
    This rule checks the location of the closing ")" character for the port map.

    The default location is on a line by itself.

    |configuring_move_token_rules_link|

    **Violation**

    .. code-block:: vhdl

        port map (
          WR_EN => wr_en);

    **Fix**

    .. code-block:: vhdl

        port map (
          WR_EN => wr_en
        );
    '''

    def __init__(self):
        Rule.__init__(self, 'port_map', '004', oToken)




from vsg import parser


class keyword(parser.keyword):
    '''
    unique_id = entity : keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class identifier(parser.identifier):
    '''
    unique_id = entity : identifier
    '''

    def __init__(self, sString):
        parser.identifier.__init__(self, sString)


class is_keyword(parser.keyword):
    '''
    unique_id = entity : is_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class begin_keyword(parser.keyword):
    '''
    unique_id = entity : begin_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class end_keyword(parser.keyword):
    '''
    unique_id = entity : end_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class end_entity_keyword(parser.keyword):
    '''
    unique_id = entity : end_entity_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class simple_name(parser.simple_name):
    '''
    unique_id = entity : simple_name
    '''

    def __init__(self, sString):
        parser.simple_name.__init__(self, sString)


class semicolon(parser.semicolon):
    '''
    unique_id = entity : semicolon
    '''

    def __init__(self):
        parser.semicolon.__init__(self)




from vsg.rules.whitespace_between_tokens import Rule

from vsg.token import loop_statement as token


class rule_101(Rule):
    '''
    This rule checks for a single space before the ending loop label if it exists.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       end loop           END_LOOP_LABEL;

    **Fix**

    .. code-block:: vhdl

       end loop END_LOOP_LABEL;
    '''
    def __init__(self):
        Rule.__init__(self, 'loop_statement', '101')
        self.left_token = token.end_loop_keyword
        self.right_token = token.end_loop_label




from vsg import parser

from vsg.vhdlFile import utils

from vsg.token import direction


def classify(iToken, lObjects):
    '''
    expression ::=
        condition_operator primary
      | logical_expression
    '''
    return utils.assign_token(lObjects, iToken, parser.todo)


def classify_until(lUntils, iToken, lObjects, oType=parser.todo):
    '''
    expression ::=
        condition_operator primary
      | logical_expression
    '''
    iCurrent = iToken
    iStop = len(lObjects) - 1
    iOpenParenthesis = 0
    iCloseParenthesis = 0
    while iCurrent < iStop:
        iCurrent = utils.find_next_token(iCurrent, lObjects)
        if utils.token_is_open_parenthesis(iCurrent, lObjects):
           iOpenParenthesis += 1
        if utils.token_is_close_parenthesis(iCurrent, lObjects):
           iCloseParenthesis += 1
#        print(f'{lObjects[iCurrent].get_value()} | {iOpenParenthesis} | {iCloseParenthesis} | {lUntils}')
        if iOpenParenthesis < iCloseParenthesis:
            break
        elif lObjects[iCurrent].get_value().lower() in lUntils:
            if utils.token_is_close_parenthesis(iCurrent, lObjects):
                if iOpenParenthesis == iCloseParenthesis:
                    utils.assign_token(lObjects, iCurrent, parser.close_parenthesis)
                    continue
                else:
                    break
            elif utils.token_is_comma(iCurrent, lObjects):
                if iOpenParenthesis == iCloseParenthesis:
                    break
                else:
                    utils.assign_token(lObjects, iCurrent, parser.comma)
            else:
                break
        else:
            utils.assign_special_tokens(lObjects, iCurrent, oType)
    return iCurrent




from vsg.token import access_type_definition as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import subtype_indication


def detect(iToken, lObjects):
    '''
    access_type_definition ::=
        access subtype_indication
    '''

    if utils.is_next_token('access', iToken, lObjects):
        return classify(iToken, lObjects)

    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('access', token.access_keyword, iToken, lObjects)

    iCurrent = subtype_indication.classify(iCurrent, lObjects)

    return iCurrent




from vsg.rules import align_tokens_in_region_between_tokens_unless_between_tokens

from vsg import token

lAlign = []
lAlign.append(token.file_declaration.colon)
lAlign.append(token.signal_declaration.colon)
lAlign.append(token.constant_declaration.colon)
lAlign.append(token.variable_declaration.colon)
lAlign.append(token.alias_declaration.colon)
lAlign.append(token.alias_declaration.is_keyword)

oStart = token.package_declaration.is_keyword
oEnd = token.package_declaration.end_keyword

lUnless = []
lUnless.append([token.subprogram_body.is_keyword,token.subprogram_body.begin_keyword])


class rule_400(align_tokens_in_region_between_tokens_unless_between_tokens):
    '''
    This rule checks the colons are in the same column for all declarations in the package declarative part.

    |configuring_keyword_alignment_rules_link|

    **Violation**

    .. code-block:: vhdl

       package my_package is

         signal   wr_en : std_logic;
         signal   rd_en   : std_logic;
         constant c_period : time;

       end package my_package;

    **Fix**

    .. code-block:: vhdl

       package my_package is

         signal   wr_en    : std_logic;
         signal   rd_en    : std_logic;
         constant c_period : time;

       end package my_package;
    '''

    def __init__(self):
        align_tokens_in_region_between_tokens_unless_between_tokens.__init__(self, 'package', '400', lAlign, oStart, oEnd, lUnless)
        self.solution = 'Align colon.'
        self.subphase = 3




from vsg.rules import insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment_and_not_on_same_line_as_token

from vsg.token import procedure_specification as token


class rule_011(insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment_and_not_on_same_line_as_token):
    '''
    This rule checks for a procedure parameter on the same line as the procedure keyword when the parameters are on multiple lines.

    **Violation**

    .. code-block:: vhdl

       procedure average_samples (constant a : in integer;
         signal d : out std_logic
       ) is
       begin


    **Fix**

    .. code-block:: vhdl

       procedure average_samples (
         constant a : in integer;
         signal d : out std_logic
       ) is
       begin
    '''

    def __init__(self):
        insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment_and_not_on_same_line_as_token.__init__(self, 'procedure', '011', token.open_parenthesis, token.close_parenthesis)
        self.solution = 'Move procedure parameter to the next line.'




from .token_indent import token_indent
from .token_indent_between_tokens import token_indent_between_tokens
from .token_indent_unless_between_tokens import token_indent_unless_between_tokens
from .remove_spaces_before_token_rule import remove_spaces_before_token_rule
from .move_token import move_token
from .move_token_next_to_another_token import move_token_next_to_another_token
from .move_token_next_to_another_token_if_it_exists_between_tokens import move_token_next_to_another_token_if_it_exists_between_tokens
from .move_token_left_to_next_non_whitespace_token import move_token_left_to_next_non_whitespace_token
from .previous_line import previous_line
from .blank_line_above_line_starting_with_token import blank_line_above_line_starting_with_token

from .line_length import line_length
from .file_length import file_length
from .token_case import token_case
from .token_case_with_prefix_suffix import token_case_with_prefix_suffix
from .token_case_in_range_bounded_by_tokens import token_case_in_range_bounded_by_tokens
from .token_case_in_range_bounded_by_tokens_with_prefix_suffix import token_case_in_range_bounded_by_tokens_with_prefix_suffix
from .blank_line_below_line_ending_with_token import blank_line_below_line_ending_with_token
from .blank_line_below_line_ending_with_several_possible_tokens import blank_line_below_line_ending_with_several_possible_tokens
from .insert_token_left_of_token_if_it_does_not_exist_between_tokens_using_value_from_token import insert_token_left_of_token_if_it_does_not_exist_between_tokens_using_value_from_token
from .insert_token_left_of_token_if_it_does_not_exist_between_tokens import insert_token_left_of_token_if_it_does_not_exist_between_tokens
from .is_token_value_one_of import is_token_value_one_of
from .align_tokens_in_region_between_tokens import align_tokens_in_region_between_tokens
from .align_tokens_in_region_between_tokens_unless_between_tokens import align_tokens_in_region_between_tokens_unless_between_tokens
from .align_tokens_in_region_between_tokens_when_between_tokens_unless_between_tokens import align_tokens_in_region_between_tokens_when_between_tokens_unless_between_tokens
from .align_tokens_in_region_between_tokens_skipping_lines_starting_with_tokens import align_tokens_in_region_between_tokens_skipping_lines_starting_with_tokens
from .align_left_token_with_right_token_if_right_token_starts_a_line import align_left_token_with_right_token_if_right_token_starts_a_line
from .remove_tokens import remove_tokens
from .remove_tokens_bounded_by_tokens_and_remove_trailing_whitespace import remove_tokens_bounded_by_tokens_and_remove_trailing_whitespace
from .insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment import insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment
from .insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment_when_between_tokens import insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment_when_between_tokens
from .insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment_and_not_on_same_line_as_token import insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment_and_not_on_same_line_as_token
from .align_consecutive_lines_starting_with_a_comment_above_line_starting_with_token import align_consecutive_lines_starting_with_a_comment_above_line_starting_with_token

from .align_consecutive_lines_after_line_starting_with_token_and_stopping_with_token import align_consecutive_lines_after_line_starting_with_token_and_stopping_with_token
from .remove_blank_lines_above_line_starting_with_token import remove_blank_lines_above_line_starting_with_token
from .blank_lines_between_token_pairs import blank_lines_between_token_pairs
from .n_spaces_before_and_after_tokens import n_spaces_before_and_after_tokens
from .spaces_before_and_after_tokens_when_bounded_by_tokens import spaces_before_and_after_tokens_when_bounded_by_tokens
from .consistent_token_case import consistent_token_case
from .consistent_case_of_tokens_from_between_tokens_applied_to_region import consistent_case_of_tokens_from_between_tokens_applied_to_region
from .token_prefix import token_prefix
from .token_prefix_between_tokens import token_prefix_between_tokens
from .token_suffix import token_suffix
from .token_suffix_between_tokens import token_suffix_between_tokens
from .split_line_at_token import split_line_at_token
from .split_line_at_token_when_between_tokens import split_line_at_token_when_between_tokens
from .split_line_at_token_when_between_tokens_unless_token_is_found import split_line_at_token_when_between_tokens_unless_token_is_found
from .insert_token_right_of_token_if_it_does_not_exist_before_token import insert_token_right_of_token_if_it_does_not_exist_before_token
from .insert_tokens_right_of_token_if_it_does_not_exist_before_token import insert_tokens_right_of_token_if_it_does_not_exist_before_token
from .insert_token_right_of_possible_tokens_if_it_does_not_exist_before_token import insert_token_right_of_possible_tokens_if_it_does_not_exist_before_token
from .remove_excessive_blank_lines_below_line_ending_with_token import remove_excessive_blank_lines_below_line_ending_with_token
from .remove_excessive_blank_lines_above_line_starting_with_token import remove_excessive_blank_lines_above_line_starting_with_token
from .formal_part_in_association_element_between_tokens import formal_part_in_association_element_between_tokens
from .move_token_sequences_left_of_token import move_token_sequences_left_of_token
from .remove_lines_starting_with_token_between_token_pairs import remove_lines_starting_with_token_between_token_pairs
from .token_case_subtype_indication import token_case_subtype_indication
from .separate_multiple_signal_identifiers_into_individual_statements import separate_multiple_signal_identifiers_into_individual_statements
from .remove_carriage_returns_between_token_pairs import remove_carriage_returns_between_token_pairs
from .split_line_at_token_if_on_same_line_as_token_if_token_pair_are_not_on_the_same_line import split_line_at_token_if_on_same_line_as_token_if_token_pair_are_not_on_the_same_line
from .remove_comments_from_end_of_lines_bounded_by_tokens import remove_comments_from_end_of_lines_bounded_by_tokens
from .token_case_formal_part_of_association_element_in_map_between_tokens import token_case_formal_part_of_association_element_in_map_between_tokens
from .token_case_n_token_after_tokens import token_case_n_token_after_tokens
from .token_case_n_token_after_tokens_between_tokens import token_case_n_token_after_tokens_between_tokens
from .existence_of_tokens_which_should_not_occur import existence_of_tokens_which_should_not_occur
from .multiline_alignment_between_tokens import multiline_alignment_between_tokens
from .multiline_structure import multiline_structure
from .multiline_simple_structure import multiline_simple_structure
from .multiline_array_alignment import multiline_array_alignment
from .multiline_constraint_structure import multiline_constraint_structure
from .multiline_procedure_call_structure import multiline_procedure_call_structure
from .multiline_conditional_alignment import multiline_conditional_alignment
from .remove_carriage_return_after_token import remove_carriage_return_after_token
from .move_token_to_the_right_of_several_possible_tokens_if_it_exists_between_tokens import move_token_to_the_right_of_several_possible_tokens_if_it_exists_between_tokens
from .number_of_lines_between_tokens import number_of_lines_between_tokens
from .token_does_not_exist_before_token import Rule as token_does_not_exist_before_token

from vsg.rules import alias_declaration
from vsg.rules import after
from vsg.rules import architecture
from vsg.rules import assert_statement
from vsg.rules import attribute
from vsg.rules import attribute_declaration
from vsg.rules import attribute_specification
from vsg.rules import block
from vsg.rules import block_comment
from vsg.rules import case
from vsg.rules import case_generate_alternative
from vsg.rules import case_generate_statement
from vsg.rules import comment
from vsg.rules import component
from vsg.rules import concurrent
from vsg.rules import constant
from vsg.rules import context
from vsg.rules import context_ref
from vsg.rules import entity
from vsg.rules import entity_specification
from vsg.rules import exit_statement
from vsg.rules import exponent
from vsg.rules import file_statement
from vsg.rules import for_loop
from vsg.rules import for_generate_statement
from vsg.rules import function
from vsg.rules import generate
from vsg.rules import generic
from vsg.rules import generic_map
from vsg.rules import ieee
from vsg.rules import if_statement
from vsg.rules import if_generate_statement
from vsg.rules import instantiation
from vsg.rules import iteration_scheme
from vsg.rules import length
from vsg.rules import library
from vsg.rules import logical_operator
from vsg.rules import loop_statement
from vsg.rules import package
from vsg.rules import package_body
from vsg.rules import port
from vsg.rules import port_map
from vsg.rules import procedure
from vsg.rules import procedure_call
from vsg.rules import process
from vsg.rules import ranges
from vsg.rules import record_type_definition
from vsg.rules import report_statement
from vsg.rules import sequential
from vsg.rules import signal
from vsg.rules import source_file
from vsg.rules import subprogram_body
from vsg.rules import subtype
from vsg.rules import type_definition
from vsg.rules import use_clause
from vsg.rules import variable
from vsg.rules import variable_assignment
from vsg.rules import wait
from vsg.rules import when
from vsg.rules import while_loop
from vsg.rules import whitespace
from vsg.rules import with_statement




from vsg import block_rule
from vsg import parser
from vsg import violation

from vsg.vhdlFile import utils


class rule_002(block_rule.Rule):
    '''
    This rule checks the **comment_left** attribute exists for all comments.

    |configuring_block_comments_link|

    **Violation**

    .. code-block:: vhdl

       --+-------------[ Header ]==============
       --   Comment
       --   Comment
       ----------------------------------------

    **Fix**

    .. code-block:: vhdl

       --+-------------[ Header ]==============
       --|  Comment
       --|  Comment
       ----------------------------------------
    '''

    def __init__(self):
        block_rule.Rule.__init__(self, 'block_comment', '002')
        self.comment_left = None
        self.configuration.append('comment_left')

    def analyze_comments(self, oToi):
        iLine, lTokens = utils.get_toi_parameters(oToi)
        iComments = utils.count_token_types_in_list_of_tokens(parser.comment, lTokens)
        iComment = 0

        for oToken in lTokens:
            iLine = utils.increment_line_number(iLine, oToken)
            iComment = utils.increment_comment_counter(iComment, oToken)

            if middle_comment(iComment, iComments, oToken):
                analyze_middle_comment(self, oToken, oToi, iLine)


def middle_comment(iComment, iComments, oToken):
    if isinstance(oToken, parser.comment):
        if iComment > 1 and iComment < iComments:
            return True
    return False


def analyze_middle_comment(self, oToken, oToi, iLine):

    self.set_token_indent(oToken)

    if self.comment_left is None:
        return None

    sHeader = self.build_comment(oToken)
    sComment = oToken.get_value()
    if not sComment.startswith(sHeader):
        sSolution = 'Comment must start with ' + sHeader
        oViolation = violation.New(iLine, oToi, sSolution)
        self.add_violation(oViolation)




from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import protected_type_body_declarative_item


def detect(iToken, lObjects):
    '''
    protected_type_body_declarative_part ::=
        { protected_type_body_declarative_item }
    '''

    return utils.detect_submodule(iToken, lObjects, protected_type_body_declarative_item)





from vsg import parser
from vsg import violation

from vsg.rules import utils as rules_utils
from vsg.rule_group import structure
from vsg.vhdlFile import utils


class move_token_left_to_next_non_whitespace_token(structure.Rule):
    '''
    Moves one token to the left until it encounters a non whitespace token.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    anchor_token : token type
       The token which another token will moved next to.

    token_to_move : token type
       The token which will be moved next to the anchor token.
    '''

    def __init__(self, name, identifier, token_to_move):
        structure.Rule.__init__(self, name, identifier)
        self.subphase = 2
        self.token_to_move = token_to_move
        self.bInsertWhitespace = True
        self.bRemoveTrailingWhitespace = True

    def _get_tokens_of_interest(self, oFile):
        lToi = oFile.get_tokens_between_non_whitespace_token_and_token(self.token_to_move)
        lReturn = []
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            if skip_based_on_whitespace(self.bInsertWhitespace, lTokens):
                continue
            lReturn.append(oToi)
        return lReturn


    def _analyze(self, lToi):
        for oToi in lToi:
            lTokens = oToi.get_tokens()

            sSolution = 'Move **then** keyword to same line as ' + lTokens[0].get_value()
            oViolation = violation.New(oToi.get_line_number(), oToi, sSolution)
            oViolation.set_remap()
            oViolation.fix_blank_lines = True
            self.add_violation(oViolation)


    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()

        rules_utils.insert_token(lTokens, 1, lTokens.pop())

        if self.bInsertWhitespace:
            rules_utils.insert_whitespace(lTokens, 1)

        lNewTokens = utils.remove_consecutive_whitespace_tokens(lTokens)

        if self.bRemoveTrailingWhitespace:
            lNewTokens = utils.remove_trailing_whitespace(lNewTokens)

        lNewTokens = utils.fix_blank_lines(lNewTokens)

        oViolation.set_tokens(lNewTokens)


def does_a_whitespace_token_separate_tokens(lTokens):
    if len(lTokens) == 3 and isinstance(lTokens[1], parser.whitespace):
        return True
    return False


def skip_based_on_whitespace(bInsertWhitespace, lTokens):
    if bInsertWhitespace and does_a_whitespace_token_separate_tokens(lTokens):
        return True
    return False




from vsg.token import record_element_constraint as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import element_constraint


def detect(iToken, lObjects):
    '''
    record_element_constraint ::=
        record_element_simple_name element_constraint
    '''
    if not utils.is_next_token('(', iToken, lObjects):
        iTemp = utils.find_next_token(iToken, lObjects) + 1
        if utils.is_next_token('(', iTemp, lObjects):
            return True
    return False


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token(token.record_element_simple_name, iToken, lObjects)
    iCurrent = element_constraint.detect(iCurrent, lObjects)

    return iCurrent




from vsg.token import subprogram_declaration as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import subprogram_specification


def detect(iToken, lObjects):
    '''
    subprogram_declaration ::=
        subprogram_specification ;
    '''

    iCurrent = subprogram_specification.detect(iToken, lObjects)
    if iCurrent != iToken:
        if utils.is_next_token(';', iCurrent, lObjects):
            return utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)
    return iCurrent




from vsg.rules.whitespace_between_tokens import Rule

from vsg.token import architecture_body as token


class rule_031(Rule):
    '''
    This rule checks for a single space between the name and the **of** keyword.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       architecture rtl    of fifo is

    **Fix**

    .. code-block:: vhdl

       architecture rtl of fifo is
    '''
    def __init__(self):
        Rule.__init__(self, 'architecture', '031')
        self.left_token = token.identifier
        self.right_token = token.of_keyword




from vsg.rules import blank_line_above_line_starting_with_token

from vsg import token

lTokens = []
lTokens.append(token.package_body.end_keyword)


class rule_202(blank_line_above_line_starting_with_token):
    '''
    This rule checks for blank lines or comments above the **end** keyword.

    |configuring_blank_lines_link|

    **Violation**

    .. code-block:: vhdl

         constant depth : integer := 512;
       end package body FIFO_PKG;

    **Fix**

    .. code-block:: vhdl

         constant depth : integer := 512;

       end package body FIFO_PKG;
    '''

    def __init__(self):
        blank_line_above_line_starting_with_token.__init__(self, 'package_body', '202', lTokens)




from vsg.rules.whitespace_before_token import Rule

from vsg.token import case_statement_alternative as token


class rule_005(Rule):
    '''
    This rule checks for a single space before the **=>** operator.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

      case data is

        when 3   =>

    **Fix**

    .. code-block:: vhdl

      case data is

        when 3 =>
    '''
    def __init__(self):
        Rule.__init__(self, 'case', '005', [token.assignment])




from vsg import token
from vsg import violation

from vsg.rule_group import case

lPortTokens = []
lPortTokens.append(token.interface_unknown_declaration.identifier)
lPortTokens.append(token.interface_constant_declaration.identifier)
lPortTokens.append(token.interface_variable_declaration.identifier)
lPortTokens.append(token.interface_signal_declaration.identifier)

lPairs = []
lPairs.append([token.concurrent_simple_signal_assignment.target, token.concurrent_simple_signal_assignment.semicolon])
lPairs.append([token.concurrent_conditional_signal_assignment.target, token.concurrent_conditional_signal_assignment.semicolon])
lPairs.append([token.constant_declaration.colon, token.constant_declaration.semicolon])
lPairs.append([token.signal_declaration.colon, token.signal_declaration.semicolon])
lPairs.append([token.variable_declaration.colon, token.variable_declaration.semicolon])
lPairs.append([token.process_statement.open_parenthesis, token.process_statement.close_parenthesis])
lPairs.append([token.process_statement.begin_keyword, token.process_statement.end_keyword])

oStartToken = token.architecture_body.entity_name
oEndToken = token.architecture_body.end_keyword


class rule_601(case.Rule):
    '''
    This rule checks for consistent capitalization of port names in an architecture body.

    **Violation**

    .. code-block:: vhdl

       entity FIFO is
         port (
           I_DATA : in std_logic_vector(31 downto 0)
         );
       end entity fifo;

       architecture rtl of fifo is

       begin

          register <= i_data;

       end architecture rtl;

    **Fix**

    .. code-block:: vhdl

       entity FIFO is
         port (
           I_DATA : in std_logic_vector(31 downto 0)
         );
       end entity fifo;

       architecture rtl of fifo is

       begin

          register <= I_DATA;

       end architecture rtl;
    '''

    def __init__(self):
        case.Rule.__init__(self, name="architecture", identifier="601")
        self.subphase = 2

    def analyze(self, oFile):
        lPorts = extract_port_names_from_entities(oFile)

        if no_ports_detected(lPorts):
            return None

        lArchitectures = oFile.get_tokens_bounded_by(oStartToken, oEndToken)
        for oArchitecture in lArchitectures:
            oArchitecture.name = extract_entity_name(oArchitecture)
            lArchitecturePorts = lPorts[oArchitecture.name]
            lToi = extract_token_pairs(oArchitecture, lPairs)
            lToi.extend(extract_component_instantiation_actual_parts(oArchitecture))
            for oToi in lToi:
                validate_port_name_in_token_list(self, lArchitecturePorts, oToi)

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dActions = oViolation.get_action()
        lTokens[0].set_value(dActions['value'])
        oViolation.set_tokens(lTokens)


def extract_component_instantiation_actual_parts(oArchitecture):
    lReturn = []

    lInstantiations = extract_component_instantiations(oArchitecture)
    for oInstantiation in lInstantiations:
        lGenericMapAspects = extract_generic_map_aspect(oInstantiation)
        lReturn.extend(extract_actual_part_tokens_from_aspects(lGenericMapAspects))
        lPortMapAspects = extract_port_map_aspect(oInstantiation)
        lReturn.extend(extract_actual_part_tokens_from_aspects(lPortMapAspects))
    return lReturn


def extract_component_instantiations(oArchitecture):
    return extract_token_pairs(oArchitecture, [[token.component_instantiation_statement.label_colon, token.component_instantiation_statement.semicolon]])


def extract_generic_map_aspect(oInstantiation):
    return extract_token_pairs(oInstantiation, [[token.generic_map_aspect.open_parenthesis, token.generic_map_aspect.close_parenthesis]])


def extract_port_map_aspect(oInstantiation):
    return extract_token_pairs(oInstantiation, [[token.port_map_aspect.open_parenthesis, token.port_map_aspect.close_parenthesis]])


def extract_actual_part_tokens_from_aspects(lMapAspect):
    lReturn = []
    for oMapAspect in lMapAspect:
        lReturn.extend(extract_actual_part_tokens_from_aspect(oMapAspect))
    return lReturn


def extract_actual_part_tokens_from_aspect(oMapAspect):
    lReturn = []
    for iToken, oToken in enumerate(oMapAspect.get_tokens()):
        lReturn.extend(extract_actual_part_token(oMapAspect, iToken, oToken))
    return lReturn


def extract_actual_part_token(oMapAspect, iToken, oToken):
    lReturn = []
    if is_actual_part_token(oToken):
        lReturn.append(oMapAspect.extract_tokens(iToken, iToken))
    return lReturn


def is_actual_part_token(oToken):
    if isinstance(oToken, token.association_element.actual_part):
        return True
    return False


def no_ports_detected(lPorts):
    if len(lPorts) == 0:
        return True
    return False


def validate_port_name_in_token_list(self, lMyPorts, oToi):
    lMyPortsLower = []
    dPortMap = {}
    for sPort in lMyPorts:
        lMyPortsLower.append(sPort.lower())
        dPortMap[sPort.lower()] = sPort

    lTokens = oToi.get_tokens()
    for iToken, oToken in enumerate(lTokens):
        sToken = oToken.get_value()
        if port_case_mismatch(sToken, lMyPorts, lMyPortsLower):
            oViolation = create_violation(sToken, iToken, dPortMap, oToi)
            self.add_violation(oViolation)


def create_violation(sToken, iToken, dPortMap, oToi):
    sSolution = 'Port case mismatch:  Change ' + sToken + ' to ' + dPortMap[sToken.lower()]
    oNewToi = oToi.extract_tokens(iToken, iToken)
    oViolation = violation.New(oNewToi.get_line_number(), oNewToi, sSolution)
    dAction = {}
    dAction['value'] = dPortMap[sToken.lower()]
    oViolation.set_action(dAction)
    return oViolation


def skip_code(oToken, bSkip):
    bReturn = bSkip

    if not bSkip:
        return start_of_skip_region_found(oToken, bReturn)

    return end_of_skip_region_found(oToken, bReturn)


def end_of_skip_region_found(oToken, bReturn):
    if isinstance(oToken, token.subprogram_body.semicolon):
        return False
    return True


def start_of_skip_region_found(oToken, bReturn):
    bReturn = function_keyword_found(oToken, bReturn)
    bReturn = procedure_keyword_found(oToken, bReturn)
    return bReturn


def function_keyword_found(oToken, bReturn):
    if isinstance(oToken, token.function_specification.function_keyword):
        return True
    return bReturn


def procedure_keyword_found(oToken, bReturn):
    if isinstance(oToken, token.procedure_specification.procedure_keyword):
        return True
    return bReturn


def port_case_mismatch(sToken, lPorts, lPortsLower):
    if sToken.lower() in lPortsLower and sToken not in lPorts:
        return True
    return False


def extract_matching_value(sFind, lValues):
    sLower = sFind.lower()
    for sValue in lValues:
        if sLower == sValue.lower():
            return sValue


def extract_token_pairs(oArchitecture, lPairs):
    lReturn = []
    for lPair in lPairs:
        lReturn.extend(extract_token_pair(oArchitecture, lPair))
    return lReturn


def extract_token_pair(oArchitecture, lPair):
    lReturn = []
    bSkip = False
    iStart = None
    for iToken, oToken in enumerate(oArchitecture.get_tokens()):
        bSkip = skip_code(oToken, bSkip)
        if bSkip:
            continue
        iStart = set_start_index(iToken, oToken, lPair, iStart)
        lReturn.extend(extract_tokens_if_end_is_found(oArchitecture, iToken, oToken, lPair, iStart))
    return lReturn


def extract_tokens_if_end_is_found(oArchitecture, iToken, oToken, lPair, iStart):
    lReturn = []
    if isinstance(oToken, lPair[1]):
        lReturn.append(oArchitecture.extract_tokens(iStart, iToken))
    return lReturn


def set_start_index(iToken, oToken, lPair, iStart):
    if isinstance(oToken, lPair[0]):
        return iToken
    return iStart


def extract_port_names_from_entities(oFile):
    lEntities = oFile.get_tokens_bounded_by(token.entity_declaration.identifier, token.entity_declaration.end_keyword)
    lPorts = {}
    for oEntity in lEntities:
        oEntity.name = extract_entity_name(oEntity)
        lPorts[oEntity.name] = extract_port_names(oEntity)
    return lPorts


def extract_entity_name(oToi):
    lTokens = oToi.get_tokens()
    for oToken in lTokens:
        if isinstance(oToken, token.entity_declaration.identifier):
            return oToken.get_value().lower()
        if isinstance(oToken, token.architecture_body.entity_name):
            return oToken.get_value().lower()


def extract_port_names(oToi):
    lReturn = []

    bSearch = False
    for oToken in oToi.get_tokens():
        bSearch = end_search(oToken, bSearch)
        lReturn.extend(extract_port_token(bSearch, oToken, lPortTokens))
        bSearch = start_search(oToken, bSearch)
    return lReturn


def end_search(oToken, bSearch):
    if isinstance(oToken, token.port_clause.close_parenthesis):
        return False
    return bSearch


def extract_port_token(bSearch, oToken, lPortTokens):
    lReturn = []
    if bSearch:
        if is_token_in_token_type_list(oToken, lPortTokens):
            lReturn.append(oToken.get_value())
    return lReturn


def start_search(oToken, bSearch):
    if isinstance(oToken, token.port_clause.open_parenthesis):
        return True
    return bSearch


def is_token_in_token_type_list(oToken, lTokenTypes):
    for oTokenType in lTokenTypes:
        if isinstance(oToken, oTokenType):
            return True
    return False




from vsg.rules import token_case_with_prefix_suffix

from vsg import token

lTokens = []
lTokens.append(token.component_declaration.identifier)


class rule_008(token_case_with_prefix_suffix):
    '''
    This rule checks the component name has proper case in the component declaration.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       component FIFO is

    **Fix**

    .. code-block:: vhdl

       component fifo is
    '''

    def __init__(self):
        token_case_with_prefix_suffix.__init__(self, 'component', '008', lTokens)
        self.groups.append('case::name')




from vsg import parser


class package_keyword(parser.keyword):
    '''
    unique_id = package_body : package_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class body_keyword(parser.keyword):
    '''
    unique_id = package_body : body_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class package_simple_name(parser.simple_name):
    '''
    unique_id = package_body : package_simple_name
    '''

    def __init__(self, sString):
        parser.simple_name.__init__(self, sString)


class is_keyword(parser.keyword):
    '''
    unique_id = package_body : is_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class end_keyword(parser.keyword):
    '''
    unique_id = package_body : end_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class end_package_keyword(parser.keyword):
    '''
    unique_id = package_body : end_package_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class end_body_keyword(parser.keyword):
    '''
    unique_id = package_body : end_body_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class end_package_simple_name(parser.simple_name):
    '''
    unique_id = package_body : end_package_simple_name
    '''

    def __init__(self, sString):
        parser.simple_name.__init__(self, sString)


class semicolon(parser.semicolon):
    '''
    unique_id = package_body : semicolon
    '''

    def __init__(self, sString):
        parser.semicolon.__init__(self)




from vsg.token import context_declaration as token

from vsg.vhdlFile.classify import context_clause

from vsg.vhdlFile import utils


def detect(iToken, lObjects):
    '''
    context_declaration ::=
        context identifier is
            context_clause
        end [ context ] [ context_simple_name ] ;
    '''

    iCurrent = utils.find_next_token(iToken, lObjects)
    if utils.object_value_is(lObjects, iCurrent, 'context'):
        if utils.find_in_range('is', iCurrent, ';', lObjects):
            return classify(iCurrent, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('context', token.context_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token(token.identifier, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required('is', token.is_keyword, iCurrent, lObjects)

    iCurrent = context_clause.detect(iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required('end', token.end_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_if('context', token.end_context_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_if_not(';', token.context_simple_name, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg import parser


def classify(lObjects, oOptions):

    if is_blank_line(lObjects, oOptions):
        lObjects.append(parser.blank_line())


def is_blank_line(lObjects, oOptions):
    if len(lObjects) == 0 and not oOptions.inside_delimited_comment():
        return True
    return False




from vsg.rules.whitespace_between_tokens import Rule as Rule

from vsg.token import record_type_definition as token


class rule_100(Rule):
    '''
    This rule checks for a single space after the **end** keyword.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       type t_record is record
         a : std_logic;
         b : std_logic;
       end         record t_record;

    **Fix**

    .. code-block:: vhdl

       type t_record is record
         a : std_logic;
         b : std_logic;
       end record t_record;
    '''
    def __init__(self):
        Rule.__init__(self, 'record_type_definition', '100')
        self.left_token = token.end_keyword
        self.right_token = token.end_record_keyword




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.concurrent_procedure_call_statement.label_name)
lTokens.append(token.procedure_call_statement.label)


class rule_300(token_indent):
    '''
    This rule checks the indent of the procedure_call label.

    **Violation**

    .. code-block:: vhdl

       a <= b;

         procedure_label : WR_EN(parameter);

    **Fix**

    .. code-block:: vhdl

       a <= b;

       procedure_label : WR_EN(parameter);
    '''

    def __init__(self):
        token_indent.__init__(self, 'procedure_call', '300', lTokens)




from vsg.rules import align_tokens_in_region_between_tokens_when_between_tokens_unless_between_tokens

from vsg import token

lAlign = []
lAlign.append(token.file_declaration.colon)
lAlign.append(token.signal_declaration.colon)
lAlign.append(token.constant_declaration.colon)
lAlign.append(token.variable_declaration.colon)
lAlign.append(token.alias_declaration.colon)
lAlign.append(token.alias_declaration.is_keyword)

oStartToken = token.for_generate_statement.generate_keyword
oEndToken = token.generate_statement_body.begin_keyword

lBetweenTokens = []
lBetweenTokens.append([token.for_generate_statement.for_keyword, token.for_generate_statement.end_keyword])

lUnless = []
lUnless.append([token.subprogram_body.is_keyword,token.subprogram_body.begin_keyword])


class rule_401(align_tokens_in_region_between_tokens_when_between_tokens_unless_between_tokens):
    '''
    This rule checks the colons are in the same column for all declarations in the generate declarative part in for generate statements.

    |configuring_keyword_alignment_rules_link|

    **Violation**

    .. code-block:: vhdl

       signal sig1: natural;
       variable var2  : natural;
       constant c_period : time;
       file my_test_input : my_file_type;

    **Fix**

    .. code-block:: vhdl

       signal sig1        : natural;
       variable var2      : natural;
       constant c_period  : time;
       file my_test_input : my_file_type;
    '''

    def __init__(self):
        align_tokens_in_region_between_tokens_when_between_tokens_unless_between_tokens.__init__(self, 'generate', '401', lAlign, oStartToken, oEndToken, lBetweenTokens, lUnless)
        self.solution = 'Align colon.'
        self.subphase = 2




from vsg import proposed_rule


class rule_028(proposed_rule.Rule):
    '''
    .. NOTE:: This rule has not been implemented yet.

    This rule checks for alignment of inline comments in the context declaration.

    |configuring_keyword_alignment_rules_link|

    **Violation**

    .. code-block:: vhdl

       context c1 is                       -- Some comment
         library ieee;                        -- Other comment
           use ieee.std_logic_1164.all;   -- Comment 3
       end context c1;  -- Comment 4

    **Fix**

    .. code-block:: vhdl

       context c1 is                    -- Some comment
         library ieee;                  -- Other comment
           use ieee.std_logic_1164.all; -- Comment 3
       end context c1;                  -- Comment 4
    '''

    def __init__(self):
        proposed_rule.Rule.__init__(self, 'context', '028')




from vsg import token

from vsg.rules import move_token_left_to_next_non_whitespace_token

oToken = token.if_statement.then_keyword


class rule_036(move_token_left_to_next_non_whitespace_token):
    '''
    This rule checks the **then** keyword is not on a line by itself.

    **Violation**

    .. code-block:: vhdl

       if a = '1'
         then

    **Fix**

    .. code-block:: vhdl

       if a = '1' then
    '''

    def __init__(self):
        move_token_left_to_next_non_whitespace_token.__init__(self, 'if', '036', oToken)





from vsg import parser

from vsg.vhdlFile.extract import tokens

from vsg.vhdlFile.extract import utils


def get_consecutive_lines_starting_with_token_and_stopping_when_token_starting_line_is_found(search_token, stop_token, lAllTokens, oTokenMap):

    lReturn = []
    lSearchIndexes = utils.filter_tokens_which_start_a_line(search_token, oTokenMap)
    lStopIndexes = utils.filter_tokens_which_start_a_line(stop_token, oTokenMap)

    lCarriageReturns = oTokenMap.get_token_indexes(parser.carriage_return)

    lSearchLines = []
    for iSearchIndex in lSearchIndexes:
        iIndex = oTokenMap.get_line_number_of_index(iSearchIndex) - 2
        lSearchLines.append(lCarriageReturns[iIndex])

    lStopLines = []
    for iStopIndex in lStopIndexes:
        iIndex = oTokenMap.get_line_number_of_index(iStopIndex) - 2
        lStopLines.append(lCarriageReturns[iIndex])

    iStart = None
    for iIndex, iValue in enumerate(lCarriageReturns):
        if iValue in lStopLines:
            if iStart is not None:
                iLine = oTokenMap.get_line_number_of_index(iStart)
                iStopIndex = lStopLines.index(iValue)
                iEnd = lStopIndexes[iStopIndex] + 1
                lTemp = lAllTokens[iStart:iEnd]
                lReturn.append(tokens.New(iStart, iLine, lTemp))
                iStart = None

        if iValue in lSearchLines:
            if iStart is None:
                iStart = iValue + 1
        else:
            iStart = None

    return lReturn




from vsg import parser
from vsg import token
from vsg import violation

from vsg.vhdlFile import utils
from vsg.rule_group import alignment
from vsg.rules import utils as rules_utils

oStart = token.process_statement.begin_keyword
oEnd = token.process_statement.end_keyword


lAssignments = []
lAssignments.append(token.simple_waveform_assignment.assignment)
lAssignments.append(token.simple_force_assignment.assignment)
lAssignments.append(token.simple_release_assignment.assignment)

lEndAssignments = []
lEndAssignments.append(token.simple_waveform_assignment.semicolon)
lEndAssignments.append(token.simple_force_assignment.semicolon)
lEndAssignments.append(token.simple_release_assignment.semicolon)


class rule_002(alignment.Rule):
    '''
    This rule checks the *after* keywords are aligned in a clock process.

    |configuring_keyword_alignment_rules_link|

    **Violation**

    .. code-block:: vhdl

       clk_proc : process(clock, reset) is
       begin
         if (reset = '1') then
           a <= '0';
           b <= '1';
         elsif (clock'event and clock = '1') then
           a <= d     after 1 ns;
           b <= c   after 1 ns;
         end if;
       end process clk_proc;

    **Fix**

    .. code-block:: vhdl

       clk_proc : process(clock, reset) is
       begin
         if (reset = '1') then
           a <= '0';
           b <= '1';
         elsif (clock'event and clock = '1') then
           a <= d     after 1 ns;
           b <= c     after 1 ns;
         end if;
       end process clk_proc;
    '''

    def __init__(self):
        alignment.Rule.__init__(self, 'after', '002')
        self.solution = 'Align **after** keyword.'
        self.disable = True
        self.subphase = 2
        self.oStart = oStart
        self.oEnd = oEnd
        self.lTokens = [token.waveform_element.after_keyword]
        ## Stuff below is from original keyword_alignment_rule

        self.compact_alignment = True
        self.configuration.append('compact_alignment')
        self.blank_line_ends_group = True
        self.configuration.append('blank_line_ends_group')
        self.comment_line_ends_group = True
        self.configuration.append('comment_line_ends_group')


    def analyze(self, oFile):
        lPreToi = oFile.get_tokens_bounded_by(self.oStart, self.oEnd)
        lToi = []

        for oToi in lPreToi:
            iLine, lTokens = utils.get_toi_parameters(oToi)
            bInsideClockDef = False
            for iToken, oToken in enumerate(lTokens):
                iLine = utils.increment_line_number(iLine, oToken)

                if not bInsideClockDef:
                    if detect_clock_definition(iToken, oToken, lTokens):
                        bInsideClockDef = True
                        iStartIndex = iToken
                    continue

                if isinstance(oToken, token.if_statement.semicolon) and oToken.get_hierarchy() == 0:
                    lToi.append(oToi.extract_tokens(iStartIndex, iToken))
                    bInsideClockDef = False
                    continue

        ### jcl - need to figure out how to do this better without copying
        dAnalysis = {}
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            iLine = oToi.get_line_number()
            iColumn = 0
            bTokenFound = False
            iToken = -1

            for iIndex in range(0, len(lTokens)):
               iToken += 1
               oToken = lTokens[iIndex]

               if not bTokenFound:
                   for oSearch in self.lTokens:
                       if isinstance(oToken, oSearch):
                           bTokenFound = True
                           dAnalysis[iLine] = {}
                           dAnalysis[iLine]['token_column'] = iColumn
                           dAnalysis[iLine]['token_index'] = iToken
                           dAnalysis[iLine]['line_number'] = iLine
                           if isinstance(lTokens[iIndex -1], parser.whitespace):
                               dAnalysis[iLine]['left_column'] = iColumn - len(lTokens[iIndex - 1].get_value())
                           else:
                               dAnalysis[iLine]['left_column'] = iColumn
                           break

                   iColumn += len(oToken.get_value())

               if isinstance(oToken, parser.carriage_return):
                   iLine += 1
                   iColumn = 0
                   bTokenFound = False
                   iToken = -1
                   if self.comment_line_ends_group:
                       if utils.are_next_consecutive_token_types([parser.whitespace, parser.comment], iIndex + 1, lTokens) or \
                          utils.are_next_consecutive_token_types([parser.comment], iIndex + 1, lTokens):
                           add_adjustments_to_dAnalysis(dAnalysis, self.compact_alignment)
                           for iKey in list(dAnalysis.keys()):
                               if dAnalysis[iKey]['adjust'] != 0:
                                   oLineTokens = oFile.get_tokens_from_line(iKey)
                                   oViolation = violation.New(oLineTokens.get_line_number(), oLineTokens, self.solution)
                                   oViolation.set_action(dAnalysis[iKey])
                                   self.add_violation(oViolation)

                           dAnalysis = {}

                   if self.blank_line_ends_group:
                       if utils.are_next_consecutive_token_types([parser.blank_line], iIndex + 1, lTokens):
                           add_adjustments_to_dAnalysis(dAnalysis, self.compact_alignment)

                           for iKey in list(dAnalysis.keys()):
                               if dAnalysis[iKey]['adjust'] != 0:
                                   oLineTokens = oFile.get_tokens_from_line(iKey)
                                   oViolation = violation.New(oLineTokens.get_line_number(), oLineTokens, self.solution)
                                   oViolation.set_action(dAnalysis[iKey])
                                   self.add_violation(oViolation)

                           dAnalysis = {}

            add_adjustments_to_dAnalysis(dAnalysis, self.compact_alignment)

            for iKey in list(dAnalysis.keys()):
                if dAnalysis[iKey]['adjust'] != 0:
                    oLineTokens = oFile.get_tokens_from_line(iKey)
                    oViolation = violation.New(oLineTokens.get_line_number(), oLineTokens, self.solution)
                    oViolation.set_action(dAnalysis[iKey])
                    self.add_violation(oViolation)

            dAnalysis = {}

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()
        iTokenIndex = dAction['token_index']

        if isinstance(lTokens[iTokenIndex - 1], parser.whitespace):
            iLen = len(lTokens[iTokenIndex - 1].get_value())
            lTokens[iTokenIndex - 1].set_value(' '*(iLen + dAction['adjust']))
        else:
            rules_utils.insert_whitespace(lTokens, iTokenIndex)
        oViolation.set_tokens(lTokens)


def add_adjustments_to_dAnalysis(dAnalysis, compact_alignment):
    iMaxLeftColumn = 0
    iMinLeftColumn = 9999999999999999
    iMaxTokenColumn = 0
    iMinTokenColumn = 9999999999999999

    for iKey in list(dAnalysis.keys()):
        iMaxLeftColumn = max(iMaxLeftColumn, dAnalysis[iKey]['left_column'])
        iMinLeftColumn = min(iMinLeftColumn, dAnalysis[iKey]['left_column'])
        iMaxTokenColumn = max(iMaxTokenColumn, dAnalysis[iKey]['token_column'])
        iMinTokenColumn = min(iMinTokenColumn, dAnalysis[iKey]['token_column'])

    if compact_alignment:
        for iKey in list(dAnalysis.keys()):
            dAnalysis[iKey]['adjust'] = iMaxLeftColumn - dAnalysis[iKey]['token_column'] + 1
    else:
        for iKey in list(dAnalysis.keys()):
            dAnalysis[iKey]['adjust'] = iMaxTokenColumn - dAnalysis[iKey]['token_column']


def detect_clock_definition(iToken, oToken, lTokens):
    if isinstance(oToken, token.if_statement.if_keyword) or isinstance(oToken, token.if_statement.elsif_keyword):
        if oToken.get_hierarchy() != 0:
            return False
        if utils.are_next_consecutive_token_types_ignoring_whitespace([parser.open_parenthesis, token.ieee.std_logic_1164.function.rising_edge], iToken + 1, lTokens) or \
           utils.are_next_consecutive_token_types_ignoring_whitespace([token.ieee.std_logic_1164.function.rising_edge], iToken + 1, lTokens) or \
           utils.are_next_consecutive_token_types_ignoring_whitespace([parser.open_parenthesis, token.ieee.std_logic_1164.function.falling_edge], iToken + 1, lTokens) or \
           utils.are_next_consecutive_token_types_ignoring_whitespace([token.ieee.std_logic_1164.function.falling_edge], iToken + 1, lTokens) or \
           utils.are_next_consecutive_token_types_ignoring_whitespace([None, parser.tic, token.predefined_attribute.event_keyword, token.logical_operator.and_operator, None, token.relational_operator.equal], iToken + 1, lTokens) or \
           utils.are_next_consecutive_token_types_ignoring_whitespace([parser.open_parenthesis, None, parser.tic, token.predefined_attribute.event_keyword, token.logical_operator.and_operator, None, token.relational_operator.equal], iToken + 1, lTokens):
            return True
    return False


def detect_signal_assignment(oToken):
    for oAssignment in lAssignments:
        if isinstance(oToken, oAssignment):
            return True
    return False


def detect_end_signal_assignment(oToken):
    for oEndAssignment in lEndAssignments:
        if isinstance(oToken, oEndAssignment):
            return True
    return False





from vsg import block_rule
from vsg import parser
from vsg import violation

from vsg.vhdlFile import utils


class rule_003(block_rule.Rule):
    '''
    This rule checks the block comment footer is correct.

    |configuring_block_comments_link|

    **Violation**

    .. code-block:: vhdl

       --+-------------[ Header ]==============
       --|  Comment
       --|  Comment
       ----------------------------------------

    **Fix**

    .. code-block:: vhdl

       --+-------------[ Header ]==============
       --|  Comment
       --|  Comment
       --+--------------------------[ Footer ]=
    '''

    def __init__(self):
        block_rule.Rule.__init__(self, 'block_comment', '003')
        self.footer_left = None
        self.footer_left_repeat = '-'
        self.footer_string = None
        self.footer_right_repeat = None
        self.footer_alignment = 'center'
        self.max_footer_column = 120
        self.configuration.extend(['footer_left', 'footer_left_repeat', 'footer_string', 'footer_right_repeat', 'footer_alignment', 'max_footer_column'])

    def analyze_comments(self, oToi):

        iLine, lTokens = utils.get_toi_parameters(oToi)
        iComments = utils.count_token_types_in_list_of_tokens(parser.comment, lTokens)

        iComment = 0
        for oToken in lTokens:
            iLine = utils.increment_line_number(iLine, oToken)
            iComment = utils.increment_comment_counter(iComment, oToken)

            if last_comment(iComment, iComments):
                analyze_footer(self, oToken, iLine, oToi)


def last_comment(iComment, iComments):
    if iComment == iComments:
        return True
    return False


def analyze_footer(self, oToken, iLine, oToi):
    sFooter = self.build_footer(oToken)
    sComment = oToken.get_value()

    if block_rule.is_footer(sComment):
        self.set_token_indent(oToken)
        if sComment != sFooter:
            sSolution = 'Change block comment footer to : ' + sFooter
            oViolation = violation.New(iLine, oToi, sSolution)
            self.add_violation(oViolation)




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.block_statement.begin_keyword)


class rule_301(token_indent):
    '''
    This rule checks the indent of the **begin** keyword.

    **Violation**

    .. code-block:: vhdl

       block_label : block is

         begin

    **Fix**

    .. code-block:: vhdl

       block_label : block is

       begin
    '''

    def __init__(self):
        token_indent.__init__(self, 'block', '301', lTokens)




from vsg.rule_group import length
from vsg import violation


class number_of_lines_between_tokens(length.Rule):
    '''
    Checks the number of lines between tokens do not exceed a specified number

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    '''

    def __init__(self, name, identifier, oLeftToken, oRightToken, iLines):
        length.Rule.__init__(self, name=name, identifier=identifier)
        self.length = iLines
        self.oLeftToken = oLeftToken
        self.oRightToken = oRightToken

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_line_count_between_tokens(self.oLeftToken, self.oRightToken)

    def _analyze(self, lToi):
        for oToi in lToi:
            if oToi.get_token_value() > self.length:
                sSolution = 'Reduce process to less than ' + str(self.length) + ' lines'
                oViolation = violation.New(oToi.get_line_number(), None, sSolution)
                self.add_violation(oViolation)




from vsg.vhdlFile.classify import context_clause
from vsg.vhdlFile.classify import library_unit


def detect(iToken, lObjects):
    '''
    design_unit ::=
        context_clause library_unit
    '''
    iCurrent = context_clause.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = library_unit.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    return iToken




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.subprogram_body.end_keyword)


class rule_003(token_indent):
    '''
    This rule checks the indent of the **end** keyword.

    **Violation**

    .. code-block:: vhdl

       procedure average_samples (
         constant a : in integer;
         signal b : in std_logic;
         variable c : in std_logic_vector(3 downto 0);
         signal d : out std_logic ) is
       begin
         end procedure average_samples;

    **Fix**

    .. code-block:: vhdl

       procedure average_samples (
         constant a : in integer;
         signal b : in std_logic;
         variable c : in std_logic_vector(3 downto 0);
         signal d : out std_logic ) is
       begin
       end procedure average_samples;
    '''

    def __init__(self):
        token_indent.__init__(self, 'procedure', '003', lTokens)




from vsg.vhdlFile.classify import simple_configuration_specification


def detect(iToken, lObjects):
    '''
    configuration_specification ::=
        simple_configuration_specification
      | compound_configuration_specification
    '''
    iReturn = simple_configuration_specification.detect(iToken, lObjects)
    if iReturn != iToken:
        return iReturn

    return iToken




from vsg import parser

from vsg.token import resolution_indication as token

from vsg.vhdlFile import utils


def detect(iToken, lObjects):
    '''
    resolution_indication ::=
        resolution_function_name | ( element_resolution )
    '''
    if detect_element_resolution(iToken, lObjects):
        return classify_element_resolution(iToken, lObjects)
    elif detect_resolution_function_name(iToken, lObjects):
        return classify_resolution_function_name(iToken, lObjects)
    return iToken


def classify_element_resolution(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('(', token.open_parenthesis, iToken, lObjects)
    iCurrent = utils.assign_tokens_until_matching_closing_paren(parser.todo, iCurrent, lObjects)

    return iCurrent


def classify_resolution_function_name(iToken, lObjects):

    return utils.assign_next_token(token.resolution_function_name, iToken, lObjects)


def detect_element_resolution(iToken, lObjects):
    if utils.is_next_token('(', iToken, lObjects):
        return True
    return False


def detect_resolution_function_name(iToken, lObjects):
    if detect_escape_value(iToken, lObjects):
        return False
    return True


lEscapeValues = ['(', ')', ';', ':=', 'range', 'bus', 'is', 'open']


def detect_escape_value(iToken, lObjects):
    iTemp = utils.find_next_token(iToken, lObjects) + 1
    if utils.is_next_token_one_of(lEscapeValues, iTemp, lObjects):
        return True
    return False




from vsg.rules import token_suffix_between_tokens

from vsg import token

lTokens = []
lTokens.append(token.interface_constant_declaration.identifier)
lTokens.append(token.interface_signal_declaration.identifier)
lTokens.append(token.interface_variable_declaration.identifier)
lTokens.append(token.interface_unknown_declaration.identifier)


class rule_025(token_suffix_between_tokens):
    '''
    This rule checks for valid suffixes on port identifiers.
    The default port suffixes are *_i*, *_o*, *_io*.

    |configuring_prefix_and_suffix_rules_link|

    **Violation**

    .. code-block:: vhdl

       port (
         wr_en    : in    std_logic;
         rd_en    : in    std_logic;
         overflow : out   std_logic;
         data     : inout std_logic
       );

    **Fix**

    .. code-block:: vhdl

       port (
         wr_en_i    : in    std_logic;
         rd_en_i    : in    std_logic;
         overflow_o : out   std_logic;
         data_io    : inout std_logic
       );
    '''

    def __init__(self):
        token_suffix_between_tokens.__init__(self, 'port', '025', lTokens, token.port_clause.open_parenthesis, token.port_clause.close_parenthesis)
        self.suffixes = ['_i', '_o', '_io']




from vsg import parser
from vsg import violation

from vsg.rule_group import blank_line

from vsg.rules import utils as rules_utils

from vsg.token import process_statement as token

from vsg.vhdlFile import utils


class rule_026(blank_line.Rule):
    '''
    This rule checks for blank lines above the first declarative line, if it exists.

    |configuring_blank_lines_link|

    **Violation**

    .. code-block:: vhdl

       proc_a : process (rd_en, wr_en, data_in, data_out,
                         rd_full, wr_full
                        ) is
         -- Keep track of the number of words in the FIFO
         variable word_count : integer;
       begin

    **Fix**

    .. code-block:: vhdl

       proc_a : process (rd_en, wr_en, data_in, data_out,
                         rd_full, wr_full
                        ) is

         -- Keep track of the number of words in the FIFO
         variable word_count : integer;
       begin
    '''

    def __init__(self):
        blank_line.Rule.__init__(self, 'process', '026')
        self.solution = 'Insert blank line below'
        self.style = 'require_blank_line'
        self.configuration.append('style')

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_tokens_bounded_by(token.process_keyword, token.begin_keyword)

    def _analyze(self, lToi):
        if self.style == 'require_blank_line':
            _analyze_require_blank_line(self, lToi)
        elif self.style == 'no_blank_line':
            _analyze_no_blank_line(self, lToi)

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()
        if dAction['action'] == 'Insert':
            rules_utils.insert_carriage_return(lTokens, dAction['index'])
            rules_utils.insert_blank_line(lTokens, dAction['index'])
            oViolation.set_tokens(lTokens)
        else:
            iStart = dAction['start']
            iEnd = dAction['end']
            lNewTokens = lTokens[:iStart]
            lNewTokens.extend(lTokens[iEnd:])
            oViolation.set_tokens(lTokens[:iStart] + lTokens[iEnd:])


def _analyze_require_blank_line(self, lToi):
    for oToi in lToi:
        lTokens = oToi.get_tokens()
        if not are_there_process_declarative_items(lTokens):
            continue
        iLine, iSearch = find_beginning_of_process_declarative_region(oToi.get_line_number(), lTokens)

        if does_a_blank_line_exist(iSearch, lTokens):
            continue

        dAction = {}
        dAction['action'] = 'Insert'
        dAction['index'] = find_carriage_return(iSearch, lTokens) + 1

        oViolation = violation.New(iLine, oToi, self.solution)
        oViolation.set_action(dAction)
        self.add_violation(oViolation)

def _analyze_no_blank_line(self, lToi):
    for oToi in lToi:
        lTokens = oToi.get_tokens()
        if not are_there_process_declarative_items(lTokens):
            continue
        iLine, iSearch = find_beginning_of_process_declarative_region(oToi.get_line_number(), lTokens)

        if not does_a_blank_line_exist(iSearch, lTokens):
            continue

        dAction = {}
        dAction['action'] = 'Remove'
        dAction['start'] = iSearch
        for iToken, oToken in enumerate(lTokens[iSearch:]):
            if isinstance(oToken, parser.carriage_return):
                if not isinstance(lTokens[iSearch + iToken + 1], parser.blank_line):
                    dAction['end'] = iSearch + iToken - 1
                    break

        oViolation = violation.New(iLine, oToi, self.solution)
        oViolation.set_action(dAction)
        self.add_violation(oViolation)

def are_there_process_declarative_items(lTokens):
    for iToken, oToken in enumerate(lTokens):

        if utils.are_next_consecutive_token_types_ignoring_whitespace([token.process_keyword, token.begin_keyword], iToken, lTokens):
            return False
        if utils.are_next_consecutive_token_types_ignoring_whitespace([token.close_parenthesis, token.begin_keyword], iToken, lTokens):
            return False
        if utils.are_next_consecutive_token_types_ignoring_whitespace([token.is_keyword, token.begin_keyword], iToken, lTokens):
            return False

    return True


def find_beginning_of_process_declarative_region(iLine, lTokens):
    iReturn = 1
    iMyLine = iLine
    iReturnLine = iLine
    for iToken, oToken in enumerate(lTokens):
        iMyLine = utils.increment_line_number(iMyLine, oToken)
        if isinstance(oToken, token.close_parenthesis):
            iReturnLine = iMyLine
            iReturn = iToken + 1
        if isinstance(oToken, token.is_keyword):
            iReturnLine = iMyLine
            iReturn = iToken + 1
            break
    return iReturnLine, iReturn


def does_a_blank_line_exist(iToken, lTokens):
    if utils.are_next_consecutive_token_types([parser.whitespace, parser.comment, parser.carriage_return, parser.blank_line], iToken, lTokens):
        return True
    if utils.are_next_consecutive_token_types([parser.comment, parser.carriage_return, parser.blank_line], iToken, lTokens):
        return True
    if utils.are_next_consecutive_token_types([parser.carriage_return, parser.blank_line], iToken, lTokens):
        return True
    return False


def find_carriage_return(iToken, lTokens):
    for iIndex in range(iToken, len(lTokens)):
        if isinstance(lTokens[iIndex], parser.carriage_return):
            return iIndex
    return None




from vsg import parser

from vsg.vhdlFile.extract import tokens
from vsg.vhdlFile.extract import utils


def get_m_tokens_before_and_n_tokens_after_token(iM, iN, lTokens, lAllTokens, oTokenMap):
    lReturn = []

    lIndexes = utils.get_indexes_of_token_list(lTokens, oTokenMap)

    for iIndex in lIndexes:
        iLine = oTokenMap.get_line_number_of_index(iIndex)
        iStart = iIndex - iM
        iEnd = iIndex + iN
        if start_index_exceeds_beginning_of_file(iStart):
            lMyTokens = []
            lMyTokens.append(parser.beginning_of_file())
            lMyTokens.extend(lAllTokens[0:iEnd + 1])
            oTokens = tokens.New(0, iLine, lMyTokens)
            lReturn.append(oTokens)
        else:
            lReturn.append(tokens.New(iStart, iLine, lAllTokens[iStart:iEnd + 1]))

    return lReturn


def start_index_exceeds_beginning_of_file(iStart):
    if iStart < 0:
        return True
    return False




from vsg import deprecated_rule


class rule_004(deprecated_rule.Rule):
    '''
    This rule has been moved to **loop_statement_103**.
    '''

    def __init__(self):
        deprecated_rule.Rule.__init__(self, 'for_loop', '004')
        self.message.append('Rule ' + self.unique_id + ' move been moved to loop_statement_103.')




from vsg.rules import previous_line

from vsg import token

lTokens = []
lTokens.append(token.if_statement.if_keyword)


class rule_031(previous_line):
    '''
    This rule checks for blank lines or comments above the **if** keyword.
    In the case of nested **if** statements, the rule will be enfoced on the first **if**.

    |configuring_previous_line_rules_link|

    The default style is :code:`no_code`.

    **Violation**

    .. code-block:: vhdl

       C <= '1';
       if (A = '1') then
         B <= '0';
       end if;

       -- This is a comment
       if (A = '1') then
         B <= '0';
       end if;

    **Fix**

    .. code-block:: vhdl

       C <= '1';

       if (A = '1') then
         B <= '0';
       end if;

       -- This is a comment
       if (A = '1') then
         B <= '0';
       end if;
    '''

    def __init__(self):
        previous_line.__init__(self, 'if', '031', lTokens)
        self.lHierarchyLimits = [0]
        self.style = 'no_code'




from .get_sequence_of_tokens_matching import get_sequence_of_tokens_matching
from .get_sequence_of_tokens_matching_bounded_by_tokens import get_sequence_of_tokens_matching_bounded_by_tokens
from .get_tokens_matching import get_tokens_matching
from .get_tokens_matching_not_at_beginning_or_ending_of_line import get_tokens_matching_not_at_beginning_or_ending_of_line
from .get_n_token_after_tokens import get_n_token_after_tokens
from .get_n_token_after_tokens_between_tokens import get_n_token_after_tokens_between_tokens
from .get_n_tokens_before_and_after_tokens import get_n_tokens_before_and_after_tokens
from .get_n_tokens_before_and_after_tokens_bounded_by_tokens import get_n_tokens_before_and_after_tokens_bounded_by_tokens
from .get_line_above_line_starting_with_token import get_line_above_line_starting_with_token
from .get_line_above_line_starting_with_token_with_hierarchy import get_line_above_line_starting_with_token_with_hierarchy
from .get_line_below_line_ending_with_token import get_line_below_line_ending_with_token
from .get_line_below_line_ending_with_several_possible_tokens import get_line_below_line_ending_with_several_possible_tokens
from .get_line_below_line_ending_with_token_with_hierarchy import get_line_below_line_ending_with_token_with_hierarchy
from .get_line_which_includes_tokens import get_line_which_includes_tokens
from .get_if_statement_conditions import get_if_statement_conditions
from .get_column_of_token_index import get_column_of_token_index
from .get_all_tokens import get_all_tokens
from .get_token_and_n_tokens_before_it import get_token_and_n_tokens_before_it
from .get_blank_lines_above_line_starting_with_token import get_blank_lines_above_line_starting_with_token
from .get_blank_lines_above_line_starting_with_token_when_between_tokens import get_blank_lines_above_line_starting_with_token_when_between_tokens
from .get_line_preceeding_line import get_line_preceeding_line
from .get_line_succeeding_line import get_line_succeeding_line
from .get_tokens_bounded_by import get_tokens_bounded_by
from .get_tokens_bounded_by_tokens_if_token_is_between_them import get_tokens_bounded_by_tokens_if_token_is_between_them
from .get_tokens_bounded_by_unless_between import get_tokens_bounded_by_unless_between
from .get_m_tokens_before_and_n_tokens_after_token import get_m_tokens_before_and_n_tokens_after_token

from .get_tokens_matching_in_range_bounded_by_tokens import get_tokens_matching_in_range_bounded_by_tokens

from .get_tokens_bounded_by_token_when_between_tokens import get_tokens_bounded_by_token_when_between_tokens
from .get_tokens_at_beginning_of_line_matching import get_tokens_at_beginning_of_line_matching
from .get_tokens_at_beginning_of_line_matching_between_tokens import get_tokens_at_beginning_of_line_matching_between_tokens
from .get_tokens_at_beginning_of_line_matching_unless_between_tokens import get_tokens_at_beginning_of_line_matching_unless_between_tokens
from .get_sequence_of_tokens_not_matching import get_sequence_of_tokens_not_matching
from .get_tokens_between_tokens_inclusive_while_storing_value_from_token import get_tokens_between_tokens_inclusive_while_storing_value_from_token
from .get_tokens_from_line import get_tokens_from_line
from .get_consecutive_lines_starting_with_token import get_consecutive_lines_starting_with_token
from .get_consecutive_lines_starting_with_token_and_stopping_when_token_starting_line_is_found import get_consecutive_lines_starting_with_token_and_stopping_when_token_starting_line_is_found
from .get_tokens_where_line_starts_with_token_until_ending_token_is_found import get_tokens_where_line_starts_with_token_until_ending_token_is_found
from .get_token_and_n_tokens_before_it_in_between_tokens import get_token_and_n_tokens_before_it_in_between_tokens
from .get_token_and_n_tokens_before_it_in_between_tokens_unless_token_is_found import get_token_and_n_tokens_before_it_in_between_tokens_unless_token_is_found
from .get_token_and_n_tokens_after_it import get_token_and_n_tokens_after_it
from .get_token_and_n_tokens_after_it_when_between_tokens import get_token_and_n_tokens_after_it_when_between_tokens
from .get_blank_lines_below_line_ending_with_token import get_blank_lines_below_line_ending_with_token
from .get_blank_lines_below_line_ending_with_several_possible_tokens import get_blank_lines_below_line_ending_with_several_possible_tokens
from .get_association_elements_between_tokens import get_association_elements_between_tokens
from .get_interface_elements_between_tokens import get_interface_elements_between_tokens
from .get_lines_with_length_that_exceed_column import get_lines_with_length_that_exceed_column
from .get_tokens_starting_with_token_and_ending_with_one_of_possible_tokens import get_tokens_starting_with_token_and_ending_with_one_of_possible_tokens
from .get_line_count_between_tokens import get_line_count_between_tokens

from .get_tokens_between_non_whitespace_token_and_token import get_tokens_between_non_whitespace_token_and_token




from vsg import parser


class keyword(parser.keyword):
    '''
    unique_id = predefined_attribute : keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class event_keyword(keyword):
    '''
    unique_id = predefined_attribute : event_keyword
    '''
    def __init__(self, sString):
        keyword.__init__(self, sString)


values = []
# Type attributes
values.append('ascending')
values.append('base')
values.append('high')
values.append('image')
values.append('left')
values.append('leftof')
values.append('low')
values.append('pos')
values.append('pred')
values.append('right')
values.append('rightof')
values.append('succ')
values.append('val')
values.append('value')

# Array attributes
values.append('ascending')
values.append('high')
values.append('left')
values.append('length')
values.append('low')
values.append('range')
values.append('reverse_range')
values.append('right')

# Signal attributes
values.append('active')
values.append('delayed')
values.append('driving')
values.append('driving_value')
values.append('event')
values.append('last_event')
values.append('last_active')
values.append('last_value')
values.append('quiet')
values.append('stable')
values.append('transaction')

# Everything attributes
values.append('instance_name')
values.append('path_name')
values.append('simple_name')





from vsg import parser
from vsg import violation

from vsg.rules import utils as rules_utils
from vsg.rule_group import alignment
from vsg.vhdlFile import utils


class align_consecutive_lines_starting_with_a_comment_above_line_starting_with_token(alignment.Rule):
    '''
    Checks for a single space between two tokens.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    token : token object
       reference token to align comments with
    '''

    def __init__(self, name, identifier, token, bIncrement=False):
        alignment.Rule.__init__(self, name=name, identifier=identifier)
        self.phase = 4
        self.subphase = 2
        self.token = token
        self.bIncrement = bIncrement

    def _get_tokens_of_interest(self, oFile):
        lToi = oFile.get_consecutive_lines_starting_with_token_and_stopping_when_token_starting_line_is_found(parser.comment, self.token)
        lReturn = []
        for oToi in lToi:
            oToi.set_meta_data('indentLevel', oFile.get_indent_of_line_at_index(oToi.get_end_index()))
            lReturn.append(oToi)

        return lReturn

    def _analyze(self, lToi):
        for oToi in lToi:
            iLine, lTokens = utils.get_toi_parameters(oToi)

            iTargetIndent = lTokens[-1].get_indent()
            sExpectedWhitespace = expected_whitespace(self, oToi)

            for iToken, oToken in enumerate(lTokens):
                iLine = utils.increment_line_number(iLine, oToken)
                if comment_at_beginning_of_line(iToken, lTokens):
                    sActualWhitespace = actual_whitespace(iToken - 1, lTokens)
                    if sExpectedWhitespace != sActualWhitespace:
                        oMyToi = extract_toi(iToken, oToi)
                        oViolation = create_violation(oMyToi, sActualWhitespace, sExpectedWhitespace, lTokens[-1])
                        self.add_violation(oViolation)
                    self._adjust_token_indent(lTokens[iToken], iTargetIndent)

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()
        if dAction['action'] == 'insert':
            rules_utils.insert_new_whitespace(lTokens, len(lTokens) - 1, dAction['whitespace'])
        else:
            lTokens[0].set_value(dAction['whitespace'])
        oViolation.set_tokens(lTokens)

    def _adjust_token_indent(self, oToken, iIndent):
        oToken.set_indent(iIndent)


def expected_whitespace(self, oToi):
    iIndentLevel = oToi.get_meta_data('indentLevel')
    if self.indentStyle == 'smart_tabs':
        return '\t' * iIndentLevel
    return ' ' * iIndentLevel * self.indentSize


def actual_whitespace(iToken, lTokens):
    if iToken == -1:
        return ''
    elif rules_utils.token_is_whitespace(lTokens[iToken]):
        return lTokens[iToken].get_value()
    return ''


def comment_at_beginning_of_line(iToken, lTokens):
    oToken = lTokens[iToken]
    if rules_utils.token_is_comment(oToken):
        if iToken == 0 or iToken == 1:
            return True
        elif rules_utils.token_is_whitespace(lTokens[iToken - 1]) and rules_utils.token_is_carriage_return(lTokens[iToken - 2]):
            return True
        elif rules_utils.token_is_carriage_return(lTokens[iToken - 1]):
            return True
    return False


def create_violation(oToi, sActualWhitespace, sExpectedWhitespace, oToken):
    iLine = oToi.get_line_number()
    sSolution = 'Align comment with ' + oToken.get_value()
    dAction = {}
    if sActualWhitespace == '':
        dAction['action'] = 'insert'
    else:
        dAction['action'] = 'adjust'
    dAction['whitespace'] = sExpectedWhitespace
    oViolation = violation.New(iLine, oToi, sSolution)
    oViolation.set_action(dAction)
    return oViolation


def extract_toi(iToken, oToi):
    lTokens = oToi.get_tokens()
    if iToken == 0:
        return oToi.extract_tokens(iToken, iToken)
    elif rules_utils.token_is_whitespace(lTokens[iToken - 1]):
        return oToi.extract_tokens(iToken - 1, iToken - 1)
    return oToi.extract_tokens(iToken, iToken)




from vsg import deprecated_rule


class rule_003(deprecated_rule.Rule):
    '''
    This rule was deprecated and replaced with rules:  function_015, package_019, procedure_010, architecture_029 and process_037.
    '''

    def __init__(self):
        deprecated_rule.Rule.__init__(self, 'constant', '003')
        self.message.append('Rule ' + self.unique_id + ' has been replaced with the following rules:')
        self.message.append('  function_015, package_019, procedure_010, architecture_029 and process_037.')




from vsg.rules import token_case

from vsg import token

lTokens = []
lTokens.append(token.attribute_declaration.attribute_keyword)


class rule_500(token_case):
    '''
    This rule checks the **attribute** keyword has proper case.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       ATTRIBUTE max_delay : time;

    **Fix**

    .. code-block:: vhdl

       attribute max_delay : time;
    '''

    def __init__(self):
        token_case.__init__(self, 'attribute_declaration', '500', lTokens)
        self.groups.append('case::keyword')




from vsg import token

from vsg.rules import token_prefix

lTokens = []
lTokens.append(token.package_body.package_simple_name)
lTokens.append(token.package_body.end_package_simple_name)


class rule_601(token_prefix):
    '''
    This rule checks for valid prefixes on package body identifiers.
    The default package prefix is *pkg_*.

    |configuring_prefix_and_suffix_rules_link|

    **Violation**

    .. code-block:: vhdl

       package body foo is

    **Fix**

    .. code-block:: vhdl

       package body pkg_foo is
    '''

    def __init__(self):
        token_prefix.__init__(self, 'package_body', '601', lTokens)
        self.prefixes = ['pkg_']
        self.solution = 'Package identifier'




from vsg import token

from vsg.rules.whitespace_between_token_pairs import Rule

lTokens = []
lTokens.append([token.function_specification.pure_keyword, token.function_specification.function_keyword])
lTokens.append([token.function_specification.impure_keyword, token.function_specification.function_keyword])
lTokens.append([token.function_specification.function_keyword, token.function_specification.designator])
lTokens.append([token.function_specification.designator, token.function_specification.return_keyword])
lTokens.append([token.function_specification.designator, token.function_specification.open_parenthesis])
lTokens.append([token.function_specification.close_parenthesis, token.function_specification.return_keyword])
lTokens.append([token.function_specification.return_keyword, token.type_mark.name])
lTokens.append([token.type_mark.name, token.subprogram_body.is_keyword])


class rule_100(Rule):
    '''
    This rule checks for a single space between the following function elements:  **function** keyword, function designator, open parenthesis, close parenthesis, **return** keyword, return type and **is** keyword.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       function     overflow    (a: integer)     return     integer    is

    **Fix**

    .. code-block:: vhdl

       function overflow (a: integer) return integer is
    '''
    def __init__(self):
        Rule.__init__(self, 'function', '100', lTokens)




import sys

from vsg import parser
from vsg import token
from vsg import violation

from vsg.vhdlFile import utils
from vsg.rule_group import structure

lIfBoundingTokens = [token.if_statement.if_keyword, token.if_statement.then_keyword]

lElsifBoundingTokens = [token.if_statement.elsif_keyword, token.if_statement.then_keyword]

oStart = token.process_statement.begin_keyword
oEnd = token.process_statement.end_keyword


class rule_029(structure.Rule):
    '''
    This rule checks for the format of clock definitions in clock processes.
    The rule can be set to enforce **event** definition:

    .. code-block:: vhdl

        if (clk'event and clk = '1') then

    ..or **edge** definition:

    .. code-block:: vhdl

        if (rising_edge(clk)) then

    event configuration
    ^^^^^^^^^^^^^^^^^^^

    .. NOTE:: This is the default configuration.

    **Violation**

    .. code-block:: vhdl

       if (rising_edge(clk)) then

       if (falling_edge(clk)) then

    **Fix**

    .. code-block:: vhdl

       if (clk'event and clk = '1') then

       if (clk'event and clk = '0') then

    edge configuration
    ^^^^^^^^^^^^^^^^^^

    .. NOTE::  Configuration this by setting the *'clock'* attribute to *'edge'*

       .. code-block:: json

          {
            "rule":{
              "process_029":{
                 "clock":"edge"
              }
            }
          }

    **Violation**

    .. code-block:: vhdl

       if (clk'event and clk = '1') then

       if (clk'event and clk = '0') then

    **Fix**

    .. code-block:: vhdl

       if (rising_edge(clk)) then

       if (falling_edge(clk)) then
    '''

    def __init__(self):
        structure.Rule.__init__(self, 'process', '029')
        self.disable = True
        self.clock = 'event'
        self.configuration.append('clock')
        self.oStart = oStart
        self.oEnd = oEnd

    def _get_tokens_of_interest(self, oFile):
        lToi = []
        aToi = oFile.get_tokens_bounded_by_token_when_between_tokens(lElsifBoundingTokens[0], lElsifBoundingTokens[1], oStart, oEnd)
        lToi = utils.combine_two_token_class_lists(lToi, aToi)
        aToi = oFile.get_tokens_bounded_by_token_when_between_tokens(lIfBoundingTokens[0], lIfBoundingTokens[1], oStart, oEnd)
        lToi = utils.combine_two_token_class_lists(lToi, aToi)
        return lToi

    def _analyze(self, lToi):
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            bEventFound = False
            iStartIndex = 0
            for iToken, oToken in enumerate(lTokens):
                if self.clock == 'edge':
                    if utils.are_next_consecutive_token_types_ignoring_whitespace([parser.tic, token.predefined_attribute.event_keyword, token.logical_operator.and_operator, None, token.relational_operator.equal], iToken + 1, lTokens):
                        bEventFound = True
                        iStartIndex = iToken
                        sSolution = 'Change event to rising_edge/falling_edge.'
                        dAction = {}
                        dAction['convert_to'] = 'edge'
                        dAction['clock'] = oToken.get_value()
                    if bEventFound and isinstance(oToken, parser.character_literal):
                        if oToken.get_value() == "'1'":
                            dAction['edge'] = 'rising_edge'
                        else:
                            dAction['edge'] = 'falling_edge'
                        oMyToi = oToi.extract_tokens(iStartIndex, iToken)
                        oViolation = violation.New(oToi.get_line_number(), oMyToi, sSolution)
                        oViolation.set_action(dAction)
                        self.add_violation(oViolation)
                        bEventFound = False

                elif self.clock == 'event':

                    if isinstance(oToken, token.ieee.std_logic_1164.function.rising_edge):
                        sSolution = 'Change rising_edge to event format.'
                        iStartIndex = iToken
                        dAction = {}
                        dAction['convert_to'] = 'event'
                        dAction['edge'] = "'1'"
                        bEventFound = True

                    elif isinstance(oToken, token.ieee.std_logic_1164.function.falling_edge):
                        sSolution = 'Change falling_edge to event format.'
                        iStartIndex = iToken
                        dAction = {}
                        dAction['convert_to'] = 'event'
                        dAction['edge'] = "'0'"
                        bEventFound = True

                    if bEventFound and isinstance(oToken, parser.todo):
                        dAction['clock'] = oToken.get_value()

                    if bEventFound and isinstance(oToken, parser.close_parenthesis):
                        oMyToi = oToi.extract_tokens(iStartIndex, iToken)
                        oViolation = violation.New(oToi.get_line_number(), oMyToi, sSolution)
                        oViolation.set_action(dAction)
                        self.add_violation(oViolation)
                        bEventFound = False
                else:
                    sys.stderr.write('Invalid configuration option ' + self.clock)
                    exit(1)

    def _fix_violation(self, oViolation):
        dAction = oViolation.get_action()
        if dAction['convert_to'] == 'edge':
            lTokens = []
            if dAction['edge'] == 'rising_edge':
                lTokens.append(token.ieee.std_logic_1164.function.rising_edge('rising_edge'))
            else:
                lTokens.append(token.ieee.std_logic_1164.function.falling_edge('falling_edge'))

            lTokens.append(parser.open_parenthesis())
            lTokens.append(parser.todo(dAction['clock']))
            lTokens.append(parser.close_parenthesis())
        else:
            lTokens = []
            lTokens.append(parser.todo(dAction['clock']))
            lTokens.append(parser.tic("'"))
            lTokens.append(token.predefined_attribute.event_keyword('event'))
            lTokens.append(parser.whitespace(' '))
            lTokens.append(token.logical_operator.and_operator('and'))
            lTokens.append(parser.whitespace(' '))
            lTokens.append(parser.todo(dAction['clock']))
            lTokens.append(parser.whitespace(' '))
            lTokens.append(token.relational_operator.equal('='))
            lTokens.append(parser.whitespace(' '))
            lTokens.append(parser.character_literal(dAction['edge']))

        oViolation.set_tokens(lTokens)




from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import design_unit


def tokenize(lObjects):
    '''
    design_file ::=
        design_unit { design_unit }
    '''
    iReturn = 0
    for iCurrent in range(0, len(lObjects)):
        iTemp = iCurrent
        if iTemp >= iReturn:
            if utils.is_item(lObjects, iTemp):
                iReturn = design_unit.detect(iTemp, lObjects)





from vsg import parser
from vsg import token
from vsg import violation
from vsg.rule_group import case
from vsg.rules import case_utils
from vsg.rules import utils


class token_case_formal_part_of_association_element_in_map_between_tokens(case.Rule):
    '''
    Checks the case for words.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    trigger : parser object type
       object type to apply the case check against
    '''

    def __init__(self, name, identifier, sMapType, oStart, oEnd):
        case.Rule.__init__(self, name=name, identifier=identifier)
        self.solution = None
        self.phase = 6
        self.case = 'lower'
        self.configuration.append('case')
        self.prefix_exceptions = []
        self.suffix_exceptions = []
        self.case_exceptions = []
        if sMapType == 'port':
            self.oMapStart = token.port_map_aspect.open_parenthesis
            self.oMapEnd = token.port_map_aspect.close_parenthesis
        else:
            self.oMapStart = token.generic_map_aspect.open_parenthesis
            self.oMapEnd = token.generic_map_aspect.close_parenthesis
        self.oStart = oStart
        self.oEnd = oEnd

    def _get_tokens_of_interest(self, oFile):
        self.case_exceptions_lower = utils.lowercase_list(self.case_exceptions)
        return oFile.get_tokens_bounded_by(self.oStart, self.oEnd)

    def _analyze(self, lToi):
        check_prefix = case_utils.is_exception_enabled(self.prefix_exceptions)
        check_suffix = case_utils.is_exception_enabled(self.suffix_exceptions)
        check_whole = case_utils.is_exception_enabled(self.suffix_exceptions)
        for oToi in lToi:
            lTokens = oToi.get_tokens()

            bMapFound = False
            bFormalFound = False
            iLine = oToi.get_line_number()
            for iToken, oToken in enumerate(lTokens):
                if isinstance(oToken, self.oMapStart):
                    bMapFound = True
                    continue
                if isinstance(oToken, self.oMapEnd):
                    bMapFound = False
                    break
                if isinstance(oToken, parser.carriage_return):
                   iLine += 1
                if isinstance(oToken, token.association_element.formal_part) and not bFormalFound and bMapFound:
                    bFormalFound = True
                    oViolation = case_utils.check_for_case_violation(oToi, self, check_prefix, check_suffix, check_whole, iToken, iLine)
                    if oViolation is not None:
                        self.add_violation(oViolation)
                if isinstance(oToken, token.association_element.assignment):
                    bFormalFound = False

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()
        lTokens[dAction['index']].set_value(dAction['value'])
        oViolation.set_tokens(lTokens)




from vsg import token

from vsg.rules import token_prefix

lTokens = []
lTokens.append(token.subtype_declaration.identifier)


class rule_004(token_prefix):
    '''
    This rule checks for valid prefixes in subtype identifiers.
    The default new subtype prefix is *st\_*.

    |configuring_prefix_and_suffix_rules_link|

    **Violation**

    .. code-block:: vhdl

       subtype my_subtype is range 0 to 9;

    **Fix**

    .. code-block:: vhdl

       subtype st_my_subtype is range 0 to 9;
    '''

    def __init__(self):
        token_prefix.__init__(self, 'subtype', '004', lTokens)
        self.prefixes = ['st_']




from vsg import token

from vsg.rules.whitespace_between_token_pairs import Rule

lTokens = []
lTokens.append([token.case_generate_statement.end_generate_keyword, token.case_generate_statement.end_generate_label])
lTokens.append([token.for_generate_statement.end_generate_keyword, token.for_generate_statement.end_generate_label])
lTokens.append([token.if_generate_statement.end_generate_keyword, token.if_generate_statement.end_generate_label])


class rule_013(Rule):
    '''
    This rule checks for a single space after the **generate** keyword and the label in the **end generate** keywords.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       end generate    ram_array;

    **Fix**

    .. code-block:: vhdl

       end generate ram_array;
    '''
    def __init__(self):
        Rule.__init__(self, 'generate', '013', lTokens)




from vsg.rules import file_length


class rule_002(file_length):
    '''
    This rule checks the length of a file.

    |configuring_length_rules_link|
    '''

    def __init__(self):
        file_length.__init__(self, 'length', '002')




from vsg.token import secondary_unit_declaration as token

from vsg.vhdlFile import utils


def detect(iToken, lObjects):
    '''
    secondary_unit_declaration ::= identifier = physical_literal;
    '''
    if utils.find_in_range('=', iToken, ';', lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):
    iCurrent = iToken
    iCurrent = utils.assign_next_token(token.identifier, iCurrent, lObjects)
    iCurrent = utils.assign_next_token(token.equal_sign, iCurrent, lObjects)

    while not utils.is_next_token(';', iCurrent, lObjects):
        iCurrent = utils.assign_next_token(token.physical_literal, iCurrent, lObjects)
    iCurrent = utils.assign_token(lObjects, iCurrent, token.semicolon)

    return iCurrent




from vsg.vhdlFile.extract import tokens


def get_tokens_matching_in_range_bounded_by_tokens(lTokens, oStart, oEnd, lAllTokens, oTokenMap):

        lStart, lEnd = oTokenMap.get_token_pair_indexes(oStart, oEnd)

        lIndexes = []
        for iStart, iEnd in zip(lStart, lEnd):
            for oToken in lTokens:
                lIndexes.extend(oTokenMap.get_token_indexes_between_indexes(oToken, iStart, iEnd))

        lIndexes.sort()

        lReturn = []
        for iIndex in lIndexes:
            iLine = oTokenMap.get_line_number_of_index(iIndex)
            lReturn.append(tokens.New(iIndex, iLine, [lAllTokens[iIndex]]))

        return lReturn



from .rule_001 import rule_001

from .rule_100 import rule_100
from .rule_101 import rule_101
from .rule_102 import rule_102

from .rule_300 import rule_300

from .rule_500 import rule_500
from .rule_501 import rule_501
from .rule_502 import rule_502
from .rule_503 import rule_503

from .rule_600 import rule_600
from .rule_601 import rule_601




from vsg.rules import token_indent_between_tokens

from vsg import token

lTokens = []
lTokens.append(token.interface_constant_declaration.constant_keyword)
lTokens.append(token.interface_signal_declaration.signal_keyword)
lTokens.append(token.interface_variable_declaration.variable_keyword)
lTokens.append(token.interface_file_declaration.file_keyword)
lTokens.append(token.interface_unknown_declaration.identifier)
lTokens.append(token.interface_incomplete_type_declaration.type_keyword)
lTokens.append(token.interface_procedure_specification.procedure_keyword)
lTokens.append(token.interface_function_specification.function_keyword)
lTokens.append(token.interface_package_declaration.package_keyword)

oStart = token.generic_clause.open_parenthesis
oEnd = token.generic_clause.close_parenthesis


class rule_004(token_indent_between_tokens):
    '''
    This rule checks the indent of generic declarations.

    **Violation**

    .. code-block:: vhdl

       generic (
       g_width : integer := 32;
              g_depth : integer := 512
       )

    **Fix**

    .. code-block:: vhdl

       generic (
         g_width : integer := 32;
         g_depth : integer := 512
       )
    '''

    def __init__(self):
        token_indent_between_tokens.__init__(self, 'generic', '004', lTokens, oStart, oEnd)




from vsg import parser


class when_keyword(parser.keyword):
    '''
    unique_id = selected_waveforms : when_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class comma(parser.comma):
    '''
    unique_id = selected_waveforms : comma
    '''

    def __init__(self, sString=','):
        parser.comma.__init__(self, sString)




from vsg.rules import insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment

from vsg.token import block_statement as token

lTokens = []
lTokens.append(token.begin_keyword)


class rule_005(insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment):
    '''
    This rule checks for code after the **begin** keyword.

    **Violation**

    .. code-block:: vhdl

       begin a <= b;

    **Fix**

    .. code-block:: vhdl

       begin
       a <= b;
    '''

    def __init__(self):
        insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment.__init__(self, 'block', '005', lTokens)
        self.solution = 'Move code after the begin to the next line.'




from vsg import token

from vsg.rules.whitespace_after_token import Rule

lTokens = []
lTokens.append(token.attribute_declaration.attribute_keyword)
lTokens.append(token.attribute_declaration.colon)


class rule_100(Rule):
    '''
    This rule checks for a single space after the following elements:  **attribute** keyword and colon.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       attribute   max_delay :   time;

    **Fix**

    .. code-block:: vhdl

       attribute max_delay : time;
    '''
    def __init__(self):
        Rule.__init__(self, 'attribute_declaration', '100', lTokens)




from vsg.rules import move_token_next_to_another_token

from vsg.token import port_map_aspect as token


class rule_003(move_token_next_to_another_token):
    '''
    This rule checks the "(" character is on the same line as the **port map** keywords.

    **Violation**

    .. code-block:: vhdl

       port map
       (
         WR_EN    => WR_EN,
         RD_EN    => RD_EN,
         OVERFLOW => OVERFLOW
       );

    **Fix**

    Use explicit port mapping.

    .. code-block:: vhdl

       port map (

         WR_EN    => WR_EN,
         RD_EN    => RD_EN,
         OVERFLOW => OVERFLOW
       );
    '''

    def __init__(self):
        move_token_next_to_another_token.__init__(self, 'port_map', '003', token.map_keyword, token.open_parenthesis)
        self.solution = 'Move the ( to the same line as the *port map* keywords.'




from vsg.token import return_statement as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import expression


def detect(iToken, lObjects):
    '''
    return_statement ::=
        [ label : ] return [ expression ] ;
    '''

    if utils.find_in_next_n_tokens(':', 2, iToken, lObjects):
        if utils.find_in_next_n_tokens('return', 3, iToken, lObjects):
            return classify(iToken, lObjects)
    if utils.is_next_token('return', iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.tokenize_label(iToken, lObjects, token.label, token.label_colon)
    iCurrent = utils.assign_next_token_required('return', token.return_keyword, iCurrent, lObjects)
    if not utils.is_next_token(';', iCurrent, lObjects):
        iCurrent = expression.classify_until([';'], iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg import violation

from vsg.rules import tokens_of_interest as toi

from vsg.vhdlFile import utils


def add_new_line(oToi):
    iToken = oToi.get_meta_data('iToken')

    oToi.set_meta_data('iStart', iToken)
    oToi.set_meta_data('sSolution', 'Move parenthesis to next line.')
    oToi.set_meta_data('sAction', 'add_new_line')
    toi.adjust_start_index_based_on_whitespace(oToi, -1)
    oViolation = _create_violation(oToi)
    return oViolation


def remove_new_line(self, oToi):
    iToken = oToi.get_meta_data('iToken')
    lTokens = oToi.get_tokens()

    oToi.set_meta_data('sSolution', 'Move parenthesis to previous line.')
    oToi.set_meta_data('sAction', 'remove_new_line')
    oToi.set_meta_data('iStart', utils.find_previous_non_whitespace_token(iToken - 1, lTokens) + 1)
    oViolation = _create_violation(oToi)
    return oViolation


def _create_violation(oToi):
    iStartIndex = oToi.get_meta_data('iStart')
    iStartLine = oToi.get_meta_data('iStartLine')
    iEndIndex = oToi.get_meta_data('iToken')
    sSolution = oToi.get_meta_data('sSolution')
    sAction = oToi.get_meta_data('sAction')
    lTokens = oToi.get_tokens()

    dAction = _create_action_dictionary(sAction)
    oViolation = violation.New(iStartLine, oToi.extract_tokens(iStartIndex, iEndIndex), sSolution)
    oViolation.set_action(dAction)
    return oViolation


def _create_action_dictionary(sAction):
    dReturn = {}
    dReturn['action'] = sAction
    return dReturn




from vsg import rule


class Rule(rule.Rule):

    def __init__(self, name, identifier):
        rule.Rule.__init__(self, name=name, identifier=identifier)
        self.fixable = False
        self.disable = True
        self.phase = 0
        self.deprecated = True
        self.message = []
        self.configuration = []

    def print_output(self):
        lReturn = [f'ERROR [config-001] Rule {self.unique_id} has been deprecated.']
        for sLine in self.message:
            lReturn.append('  ' + sLine)
        return lReturn




from vsg.token import procedure_call_statement as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import procedure_call


lKeywords = ['null', 'return', 'exit', 'next', 'while', 'for', 'loop', 'case', 'if', 'report', 'assert', 'wait', 'end', 'with', 'else', 'elsif', 'when']


def detect(iToken, lObjects):
    '''
    procedure_call_statement ::=
        [ label : ] procedure_call ;
    '''

    iCurrent = iToken
    # Move past label if it exists
    if utils.find_in_next_n_tokens(':', 2, iCurrent, lObjects):
        iCurrent = utils.find_next_token(iCurrent, lObjects)
        iCurrent +=1
        iCurrent = utils.find_next_token(iCurrent, lObjects)
        iCurrent +=1
    # Check if next token is keyword
    iCurrent = utils.find_next_token(iCurrent, lObjects)
    if lObjects[iCurrent].get_value().lower() in lKeywords:
        return iToken
    # Check if signal assignment operator exists
    if not utils.all_assignments_inside_parenthesis(iToken, ';', lObjects):
        return iToken
    # Check if variable assignment operator exists
    if utils.find_in_range(':=', iCurrent, ';', lObjects):
        return iToken
    # Otherwise it must be a procedure_call_statement
    return classify(iToken, lObjects)


def classify(iToken, lObjects):

    iCurrent = utils.tokenize_label(iToken, lObjects, token.label, token.label_colon)

    iCurrent = procedure_call.classify(iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg.rules import token_indent as Rule

from vsg.token import record_type_definition as token


class rule_301(Rule):
    '''
    This rule checks the indent of the **end** keyword.

    **Violation**

    .. code-block:: vhdl

       type t_record is record
         a : std_logic;
         b : std_logic;
           end record t_record;

    **Fix**

    .. code-block:: vhdl

       type t_record is record
         a : std_logic;
         b : std_logic;
       end record t_record;
    '''

    def __init__(self):
        Rule.__init__(self, 'record_type_definition', '301', [token.end_keyword])




from vsg.rules import token_suffix_between_tokens

from vsg import token

lTokens = []
lTokens.append(token.interface_constant_declaration.identifier)
lTokens.append(token.interface_signal_declaration.identifier)
lTokens.append(token.interface_variable_declaration.identifier)
lTokens.append(token.interface_unknown_declaration.identifier)


class rule_600(token_suffix_between_tokens):
    '''
    This rule checks for valid suffixes on generic identifiers.
    The default generic suffix is *\_g*.

    |configuring_prefix_and_suffix_rules_link|

    **Violation**

    .. code-block:: vhdl

       generic(my_generic : integer);

    **Fix**

    .. code-block:: vhdl

       generic(my_generic_g : integer);
    '''

    def __init__(self):
        token_suffix_between_tokens.__init__(self, 'generic', '600', lTokens, token.generic_clause.open_parenthesis, token.generic_clause.close_parenthesis)
        self.suffixes = ['_g']




from vsg import parser
from vsg import token

from vsg.rules import consistent_token_case

lTokens = []
lTokens.append(token.incomplete_type_declaration.identifier)
lTokens.append(token.full_type_declaration.identifier)

lIgnore = []
lIgnore.append(token.interface_signal_declaration.identifier)
lIgnore.append(token.interface_unknown_declaration.identifier)
lIgnore.append(token.interface_constant_declaration.identifier)
lIgnore.append(token.interface_variable_declaration.identifier)
lIgnore.append(token.association_element.formal_part)
lIgnore.append(parser.whitespace)
lIgnore.append(parser.carriage_return)
lIgnore.append(parser.blank_line)


class rule_014(consistent_token_case):
    '''
    This rule checks for consistent capitalization of type names.

    **Violation**

    .. code-block:: vhdl

       type state_machine is (idle, write, read, done);

       signal sm : State_Machine;

    **Fix**

    .. code-block:: vhdl

       type state_machine is (idle, write, read, done);

       signal sm : state_machine;
    '''

    def __init__(self):
        consistent_token_case.__init__(self, 'type', '014', lTokens, lIgnore)




from vsg import parser
from vsg import token
from vsg import violation

from vsg.rules import alignment_utils
from vsg.rules import utils as rules_utils
from vsg.rule_group import alignment
from vsg.vhdlFile import utils


class align_tokens_in_region_between_tokens(alignment.Rule):
    '''
    Checks for a single space between two tokens.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    lTokens : token object list
       List of tokens to align

    left_token : token object
       The first token that defines the region

    right_token : token object
       The second token that defines the region
    '''

    def __init__(self, name, identifier, lTokens, left_token, right_token, bIndexes=False):
        alignment.Rule.__init__(self, name=name, identifier=identifier)
        self.lTokens = lTokens
        self.left_token = left_token
        self.right_token = right_token
        self.bIndexes = bIndexes
        ## Stuff below is from original keyword_alignment_rule

        self.compact_alignment = True
        self.configuration.append('compact_alignment')

        self.blank_line_ends_group = True
        self.configuration.append('blank_line_ends_group')
        self.comment_line_ends_group = True
        self.configuration.append('comment_line_ends_group')
        self.separate_generic_port_alignment = True
        self.configuration.append('separate_generic_port_alignment')
        self.generate_statement_ends_group = False
#        self.configuration.append('generate_statement_ends_group')
        self.bIncludeTillBeginningOfLine = False

    def analyze(self, oFile):
        lToi = oFile.get_tokens_bounded_by(self.left_token, self.right_token, bIncludeTillBeginningOfLine=self.bIncludeTillBeginningOfLine)
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            iLine = oToi.get_line_number()
            iColumn = 0
            bTokenFound = False
            iToken = -1
            dAnalysis = {}

            if rules_utils.number_of_carriage_returns(lTokens) == 0:
                continue

            for iIndex in range(0, len(lTokens)):
               iToken += 1
               oToken = lTokens[iIndex]

               if not bTokenFound:
                   for oSearch in self.lTokens:
                       if isinstance(oToken, oSearch):
                           bTokenFound = True
                           dAnalysis[iLine] = {}
                           dAnalysis[iLine]['token_column'] = iColumn
                           dAnalysis[iLine]['token_index'] = iToken
                           dAnalysis[iLine]['line_number'] = iLine
                           dAnalysis[iLine]['token_value'] = oToken.get_value()
                           if isinstance(lTokens[iIndex -1], parser.whitespace):
                               dAnalysis[iLine]['left_column'] = iColumn - len(lTokens[iIndex - 1].get_value())
                           else:
                               dAnalysis[iLine]['left_column'] = iColumn
                           break

                   iColumn += alignment_utils.update_column_width(self, oToken)

               if isinstance(oToken, token.generic_clause.semicolon) and self.separate_generic_port_alignment:
                   add_adjustments_to_dAnalysis(dAnalysis, self.compact_alignment)
                   for iKey in list(dAnalysis.keys()):
                       if dAnalysis[iKey]['adjust'] != 0:
                           oLineTokens = oFile.get_tokens_from_line(iKey)
                           sSolution = 'Move ' + dAnalysis[iKey]['token_value'] + ' ' + str(dAnalysis[iKey]['adjust']) + ' columns'
                           oViolation = violation.New(oLineTokens.get_line_number(), oLineTokens, sSolution)
                           oViolation.set_action(dAnalysis[iKey])
                           self.add_violation(oViolation)

                   dAnalysis = {}

               if isinstance(oToken, token.generic_map_aspect.close_parenthesis) and self.separate_generic_port_alignment:
                   add_adjustments_to_dAnalysis(dAnalysis, self.compact_alignment)
                   for iKey in list(dAnalysis.keys()):
                       if dAnalysis[iKey]['adjust'] != 0:
                           oLineTokens = oFile.get_tokens_from_line(iKey)
                           sSolution = 'Move ' + dAnalysis[iKey]['token_value'] + ' ' + str(dAnalysis[iKey]['adjust']) + ' columns'
                           oViolation = violation.New(oLineTokens.get_line_number(), oLineTokens, sSolution)
                           oViolation.set_action(dAnalysis[iKey])
                           self.add_violation(oViolation)

                   dAnalysis = {}

               if generate_statement_detected(self, oToken):
                   add_adjustments_to_dAnalysis(dAnalysis, self.compact_alignment)
                   for iKey in list(dAnalysis.keys()):
                       if dAnalysis[iKey]['adjust'] != 0:
                           oLineTokens = oFile.get_tokens_from_line(iKey)
                           sSolution = 'Move ' + dAnalysis[iKey]['token_value'] + ' ' + str(dAnalysis[iKey]['adjust']) + ' columns'
                           oViolation = violation.New(oLineTokens.get_line_number(), oLineTokens, sSolution)
                           oViolation.set_action(dAnalysis[iKey])
                           self.add_violation(oViolation)

                   dAnalysis = {}

               if isinstance(oToken, parser.carriage_return):
                   iLine += 1
                   iColumn = 0
                   bTokenFound = False
                   iToken = -1
                   if self.comment_line_ends_group:
                       if utils.are_next_consecutive_token_types([parser.whitespace, parser.comment], iIndex + 1, lTokens) or \
                          utils.are_next_consecutive_token_types([parser.comment], iIndex + 1, lTokens):
                           add_adjustments_to_dAnalysis(dAnalysis, self.compact_alignment)
                           for iKey in list(dAnalysis.keys()):
                               if dAnalysis[iKey]['adjust'] != 0:
                                   oLineTokens = oFile.get_tokens_from_line(iKey)
                                   sSolution = 'Move ' + dAnalysis[iKey]['token_value'] + ' ' + str(dAnalysis[iKey]['adjust']) + ' columns'
                                   oViolation = violation.New(oLineTokens.get_line_number(), oLineTokens, sSolution)
                                   oViolation.set_action(dAnalysis[iKey])
                                   self.add_violation(oViolation)

                           dAnalysis = {}

                   if self.blank_line_ends_group:
                       if utils.are_next_consecutive_token_types([parser.blank_line], iIndex + 1, lTokens):
                           add_adjustments_to_dAnalysis(dAnalysis, self.compact_alignment)

                           for iKey in list(dAnalysis.keys()):
                               if dAnalysis[iKey]['adjust'] != 0:
                                   oLineTokens = oFile.get_tokens_from_line(iKey)
                                   sSolution = 'Move ' + dAnalysis[iKey]['token_value'] + ' ' + str(dAnalysis[iKey]['adjust']) + ' columns'
                                   oViolation = violation.New(oLineTokens.get_line_number(), oLineTokens, sSolution)
                                   oViolation.set_action(dAnalysis[iKey])
                                   self.add_violation(oViolation)

                           dAnalysis = {}

            add_adjustments_to_dAnalysis(dAnalysis, self.compact_alignment)

            for iKey in list(dAnalysis.keys()):
                if dAnalysis[iKey]['adjust'] != 0:
                    oLineTokens = oFile.get_tokens_from_line(iKey)
                    sSolution = 'Move ' + dAnalysis[iKey]['token_value'] + ' ' + str(dAnalysis[iKey]['adjust']) + ' columns'
                    oViolation = violation.New(oLineTokens.get_line_number(), oLineTokens, sSolution)
                    oViolation.set_action(dAnalysis[iKey])
                    self.add_violation(oViolation)

            dAnalysis = {}

    def _fix_violation(self, oViolation):

        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()
        iTokenIndex = dAction['token_index']

        if isinstance(lTokens[iTokenIndex - 1], parser.whitespace):
            iLen = len(lTokens[iTokenIndex - 1].get_value())
            lTokens[iTokenIndex - 1].set_value(' '*(iLen + dAction['adjust']))
        else:
            rules_utils.insert_whitespace(lTokens, iTokenIndex, dAction['adjust'])
        oViolation.set_tokens(lTokens)


def add_adjustments_to_dAnalysis(dAnalysis, compact_alignment):
    iMaxLeftColumn = 0
    iMinLeftColumn = 9999999999999999
    iMaxTokenColumn = 0
    iMinTokenColumn = 9999999999999999

    for iKey in list(dAnalysis.keys()):
        iMaxLeftColumn = max(iMaxLeftColumn, dAnalysis[iKey]['left_column'])
        iMinLeftColumn = min(iMinLeftColumn, dAnalysis[iKey]['left_column'])
        iMaxTokenColumn = max(iMaxTokenColumn, dAnalysis[iKey]['token_column'])
        iMinTokenColumn = min(iMinTokenColumn, dAnalysis[iKey]['token_column'])

    if compact_alignment:
        for iKey in list(dAnalysis.keys()):
            dAnalysis[iKey]['adjust'] = iMaxLeftColumn - dAnalysis[iKey]['token_column'] + 1
    else:
        for iKey in list(dAnalysis.keys()):
            dAnalysis[iKey]['adjust'] = iMaxTokenColumn - dAnalysis[iKey]['token_column']


def generate_statement_detected(self, oToken):
    if not self.generate_statement_ends_group:
        return False
    if generate_label_detected(oToken):
        return True
    if generate_semicolon_detected(oToken):
        return True
    if case_generate_alternative_detected(oToken):
        return True
    return False


def generate_label_detected(oToken):
    if isinstance(oToken, token.if_generate_statement.generate_label):
        return True
    if isinstance(oToken, token.for_generate_statement.generate_label):
        return True
    return False


def generate_semicolon_detected(oToken):
    if isinstance(oToken, token.if_generate_statement.semicolon):
        return True
    if isinstance(oToken, token.for_generate_statement.semicolon):
        return True
    if isinstance(oToken, token.case_generate_statement.semicolon):
        return True
    return False


def case_generate_alternative_detected(oToken):
    if isinstance(oToken, token.case_generate_alternative.when_keyword):
        return True
    return False




from vsg import parser


class begin_keyword(parser.keyword):
    '''
    unique_id = generate_statement_body : begin_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class end_keyword(parser.keyword):
    '''
    unique_id = generate_statement_body : end_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class alternative_label(parser.label):
    '''
    unique_id = generate_statement_body : alternative_label
    '''

    def __init__(self, sString):
        parser.label.__init__(self, sString)


class semicolon(parser.semicolon):
    '''
    unique_id = generate_statement_body : semicolon
    '''

    def __init__(self, sString=';'):
        parser.semicolon.__init__(self)




from vsg.token import protected_type_declaration as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import protected_type_declarative_part


def detect(iToken, lObjects):
    '''
    protected_type_declaration ::=
        **protected**
        protected_type_declarative_part
        **end** **protected** [ protected_type_simple_name ]
    '''

    if utils.is_next_token('protected', iToken, lObjects):
        if not utils.are_next_consecutive_tokens(['protected', 'body'], iToken, lObjects):
            return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('protected', token.protected_keyword, iToken, lObjects)

    iCurrent = protected_type_declarative_part.detect(iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required('end', token.end_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token_required('protected', token.end_protected_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token_if_not(';', token.protected_type_simple_name, iCurrent, lObjects)

    return iCurrent




from vsg.rules import remove_excessive_blank_lines_below_line_ending_with_token as Rule

from vsg.token import entity_declaration as token


class rule_201(Rule):
    '''
    This rule ensures no blank lines after the **is** keyword.

    **Violation**

    .. code-block:: vhdl

       entity fifo is


         port (

    **Fix**

    .. code-block:: vhdl

       entity fifo is
         port (
    '''
    def __init__(self):
        Rule.__init__(self, 'entity', '201', [token.is_keyword], iAllow=0)




from vsg import token

from vsg.rules import token_suffix

lTokens = []
lTokens.append(token.component_instantiation_statement.instantiation_label)


class rule_600(token_suffix):
    '''
    This rule checks for valid suffixes on instantiation labels.
    The default suffix is *_inst*.

    |configuring_prefix_and_suffix_rules_link|

    **Violation**

    .. code-block:: vhdl

       fifo_32x2k : FIFO

    **Fix**

    .. code-block:: vhdl

       fifo_32x2k_inst : FIFO
    '''

    def __init__(self):
        token_suffix.__init__(self, 'instantiation', '600', lTokens)
        self.suffixes = ['_inst']
